

def plot(row, treeCol, muCol, Ts, eventAttr, leafLabelAttr, args):
    ax= plot_recon(row[treeCol], Ts, row[muCol],
                   eventAttr= eventAttr,
                   eventSymbols= {'S':'.r', 'D':'db'},
                   sColor= '#97a4bd',
                   gLineStyle= '-k',
                   sRoot= 0,
                   leafLabelAttr= leafLabelAttr,
                   leafLabelStyle= {'fontsize': 10, 'color': '#B37700'},
                   pipe_r= 2,
                   ax= None,
                   )
    SAVE(ax, row, args)
    return None

def SAVE(ax, row, args):
    opath1= f"{args.save_to}{row.OG}.{args.format}"
    ax.figure.savefig(opath1)
    print(f'Saved to {opath1}')


desc_def= """
"""


if __name__ == "__main__":
    from .plot_functions import plot_recon
    from .nhx_tools import read_nhx
    from .common_tools import norm_path, txtMu_2_dict, update_mu, add_eventAttr

    import argparse
    from importlib.metadata import version
    from pandas import read_csv
    from tqdm import tqdm
    tqdm.pandas()

    V_tl= version('revolutionhtl')
    txt= f'REvolutionH-tl: Reconstruction of Evolutionaty Histories TooL (V{V_tl})'

    parser = argparse.ArgumentParser(prog= 'revolutionhtl',
                                     description=f'{txt}{desc_def}',
                                     usage='python -m revolutionhtl <arguments>',
                                     formatter_class=argparse.MetavarTypeHelpFormatter,
                                    )

    #############
    # Arguments #
    #############

    parser.add_argument('OGs_mask',
                        help= '[str] List of orthogroups IDs to mask gene trees at the file "gene_trees"',
                        type= str,
                        nargs= '*',
                        default= None,
                       )

    parser.add_argument('-files_path',
                        help= '[str] Path of files outpued by REvolutionH-tl.',
                        type= norm_path,
                        #nargs='?',
                        default= './',
                       )

    parser.add_argument('--files_prefix',
                        help= '[str] Prefix of files outpued by REvolutionH-tl.',
                        type= str,
                        nargs='?',
                        default= 'tl_project.',
                       )

    parser.add_argument('--save_to',
                        help= '[str] Directory to save figures. If not specified, they will be displayed in the sceen.',
                        type= norm_path,
                        default= './',
                       )

    parser.add_argument('--format',
                        help= '[str | default: svg] Format to save figure',
                        type= str,
                        default= 'svg',
                       )

    args= parser.parse_args()
    ################
    # Process data #
    ################

    from .hello import hello5
    print(f'{hello5}V{V_tl}\n')

    prefix= f'{args.files_path}{args.files_prefix}'
    args.species_tree= f'{prefix}labeled_species_tree.nhx'
    args.gene_trees=   f'{prefix}reconciliation.tsv'
    args.tree_column= 'tree'
    args.mu_column= 'reconciliation_map'
    args.OG_column= 'OG'
    args.event_attr= 'label'
    args.label_attr= 'label'
    args.node_id= 'node_id'

    # Load data
    #----------

    # Read species tree
    with open(args.species_tree) as F:
        Ts= read_nhx(''.join( F.read().strip().split('\n') ))
    # Read gene trees table
    df= read_csv(args.gene_trees, sep= '\t',  dtype= {args.OG_column:str})
    # Mask gene trees
    if args.OGs_mask:
        df= df[ df[args.OG_column].isin(args.OGs_mask) ]
    else:
        raise ValueError('Missing parameter: OGs_mask (list of orthogroup IDs to plot).')
    # Read gene trees
    df[args.tree_column]= df[args.tree_column].apply(read_nhx)
    # Set mu map in proper format
    df[args.mu_column]= df[args.mu_column].apply(txtMu_2_dict)
    # Set mu with nx node identifier
    df[args.mu_column]= df.apply(lambda row: update_mu(row[args.tree_column],
                                                    Ts, row[args.mu_column],
                                                    args.node_id, args.node_id),
                                 axis=1)
    df[args.mu_column].apply(lambda D: D.update({0:0}))
    # Add custom event attr
    customEventAttr= 'customEventAttr'
    F= lambda T: add_eventAttr(T, args.event_attr, customEventAttr)
    df[args.tree_column].apply(F)

    # Plot
    #-----
    F= lambda row: plot(row, args.tree_column, args.mu_column,
                        Ts, customEventAttr, args.label_attr, args)

    df.apply(F, axis= 1)



'''This file was generated automatically by `r_pythonize.py`.
It converts R modules into a pandas series,
ready to be called by rpy2.robjects.r() method.
Author of R code:
Gabriel Martinez [gabriel.martinez@cinvestav.mx]
Andres Tinajero [jesus.tinajero@cinvestav.mx]'''

from pandas import Series
modules= dict()
modules['Module_1']= '''########################################
# Module 1 | Install and load packages #
########################################


## Step 0: Load and install required packages.

# Set the CRAN repository at the beginning
options(repos = c(CRAN = "https://cran.r-project.org"))

# Install BiocManager if it is not already available
if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

# Set Bioconductor to the desired version
BiocManager::install(version = "3.20", ask = FALSE)

# Function to check and install packages from CRAN and Bioconductor
check_and_install_packages <- function(packages) {
  # Initialize lists to store installation and loading results
  installed_successfully <- c()
  loaded_successfully <- c()
  failed_to_install <- c()
  failed_to_load <- c()

  # Identify packages that are not installed
  packages_to_install <- packages[!(packages %in% installed.packages()[, "Package"])]

  # Specific Bioconductor packages
  bioc_packages <- c("ggtree", "treeio", "ggtreeExtra", "phangorn")

  # Separate packages by source: CRAN or Bioconductor
  cran_packages <- setdiff(packages_to_install, bioc_packages)
  bioc_packages_to_install <- intersect(packages_to_install, bioc_packages)

  # Install packages from CRAN
  if (length(cran_packages) > 0) {
    tryCatch({
      install.packages(cran_packages, dependencies = TRUE)
      installed_successfully <- c(installed_successfully, cran_packages)
    }, error = function(e) {
      message("Error installing packages from CRAN: ", conditionMessage(e))
      failed_to_install <- c(failed_to_install, cran_packages)
    })
  }

  # Install packages from Bioconductor
  if (length(bioc_packages_to_install) > 0) {
    tryCatch({
      BiocManager::install(bioc_packages_to_install, ask = FALSE, update = FALSE)
      installed_successfully <- c(installed_successfully, bioc_packages_to_install)
    }, error = function(e) {
      message("Error installing packages from Bioconductor: ", conditionMessage(e))
      failed_to_install <- c(failed_to_install, bioc_packages_to_install)
    })
  }

  # Load the installed packages
  lapply(packages, function(pkg) {
    if (require(pkg, character.only = TRUE)) {
      loaded_successfully <<- c(loaded_successfully, pkg)
    } else {
      failed_to_load <<- c(failed_to_load, pkg)
    }
  })

  # Display a summary of installation and loading results
  message("\n--- Package Installation and Loading Summary ---")
  if (length(installed_successfully) > 0) {
    message("Successfully installed packages: ", paste(installed_successfully, collapse = ", "))
  }
  if (length(failed_to_install) > 0) {
    message("Packages that failed to install: ", paste(failed_to_install, collapse = ", "))
  }
  if (length(loaded_successfully) > 0) {
    message("Successfully loaded packages: ", paste(loaded_successfully, collapse = ", "))
  }
  if (length(failed_to_load) > 0) {
    stop("Packages that failed to load: ", paste(failed_to_load, collapse = ", "))
  }
}

# List of required packages
required_packages <- c("ggtree", "tidytree", "ggplot2", "dplyr", "tidyr",
                       "RColorBrewer", "gridExtra", "cowplot", "treeio",
                       "ggtreeExtra", "phangorn", "grid", "magick", "here",
                       "ggimage", "svglite")

# Check and install the required packages
check_and_install_packages(required_packages)

'''
modules['Module_2']= '''#######################################
# Module 2 | Read, load and plot tree #
#######################################

# This script will be called by the python program revolutionhtl.plot_summary
# The variables bellow are suposed to be created by python, to reset uncoment lines below
#species_tree_file <- "tl_project.labeled_species_tree.nhxx.corrected"
#numbers_file <- "numbers_reconciliation.tsv"

## Step 1: Read and process your tree
newick_string <- paste(readLines(species_tree_file), collapse = "")
tree <- read.nhx(textConnection(newick_string))

# Reading and combining datan
Datos <- read.table(numbers_file, header = TRUE)
#Datos <- Datos %>%
# rename(node = node_id)
Datos <- Datos %>%
  mutate(Speciation_genes = speciation_roots + genes_at_speciation) %>%
  mutate(Gainss = speciation_roots + duplication_roots)

# Merge additional data with tree data based on the node ID
tree_data <- tree@data
tree_data_combined <- left_join(tree_data, Datos, by = "node_id")
tree@data <- tree_data_combined

# Function to calculate the maximum depth of internal nodes in the tree
calculate_max_depth_internal <- function(tree) {
  # Convert the `treedata` object to `phylo` if needed
  if (inherits(tree, "treedata")) {
    tree <- as.phylo(tree)
  }

  num_tips <- length(tree$tip.label)
  edge_matrix <- tree$edge

  # Helper function to calculate node depth recursively
  calculate_depth <- function(node, tree_edges, num_tips) {
    # If the node is a tip, the depth is 0
    if (node <= num_tips) {
      return(0)
    }

    # Find child nodes of the current node
    child_nodes <- tree_edges[tree_edges[, 1] == node, 2]

    # Calculate the depth of each child and return the maximum
    depths <- sapply(child_nodes, function(child) calculate_depth(child, tree_edges, num_tips))

    # Return the maximum depth of children + 1 for the current node
    return(max(depths) + 1)
  }

  # Calculate the maximum depth of all internal nodes (non-tips)
  depths <- sapply((num_tips + 1):(num_tips + tree$Nnode), function(node) calculate_depth(node, edge_matrix, num_tips))

  # Return the maximum depth of internal nodes
  return(max(depths) - 1)
}

# Function to calculate the number of species (tips) in the tree
calculate_num_species <- function(tree) {
  # Convert the `treedata` object to `phylo` if needed
  if (inherits(tree, "treedata")) {
    tree <- as.phylo(tree)
  }

  # The number of species is the number of tips (leaves) in the tree
  return(length(tree$tip.label))
}

# Calculate the number of species and the maximum depth of the tree
num_species <- as.numeric(calculate_num_species(tree))
max_depth <- as.numeric(calculate_max_depth_internal(tree))

# Predefined data for node-depth limits
limits_table <- data.frame(
  Nodos = 0:14,
  Limits = c(1.2, 2.5, 3.5, 4.7, 5.7, 6.7, 8, 9, 10, 11, 12, 13, 14, 15.3, 16.3)
)

# Fit a linear regression model for dynamic x-axis limits
lm_model <- lm(Limits ~ Nodos, data = limits_table)

# Function to calculate the x-axis limit dynamically based on max depth
get_plot_limit <- function(max_depth) {
  if (max_depth >= 0) {  # Ensure max_depth is valid
    # Predict the limit using the regression model
    limit_value <- predict(lm_model, newdata = data.frame(Nodos = max_depth))
    return(limit_value)
  } else {
    stop("max_depth must be greater than or equal to 0.")
  }
}

# Get the dynamic plot limit
plot_limit <- get_plot_limit(max_depth)

# Create the tree plot with ggtree
p <- ggtree(tree, ladderize = FALSE) +
  geom_tree(size = 1) +  # Draw the tree structure
  geom_tiplab(size = 4, vjust = 0.5) +  # Add labels for the tips
  geom_text(aes(x = branch, label = Speciation_genes), size = 4, vjust = -0.2, hjust = 1.3, color = "black") +  # Add branch labels
  scale_x_continuous(limits = c(0, plot_limit)) +  # Set the x-axis limit dynamically
  coord_cartesian(ylim = c(1, ifelse(num_species > 20, num_species + 1, num_species))) +  # Adjust y-axis limit based on species count
  theme(
    panel.grid = element_blank()  # Remove grid lines from the plot
  )

#################################################################################
'''
modules['Module_3']= '''############################################
# Module 3 | Calculations and prepare data #
############################################

# This script will be called by the python program revolutionhtl.plot_summary

# Calculation of gains, losses, and duplications

tree_data <- tree_data_combined %>%
  mutate(
    total = gene_gain_by_duplication + gene_loss + duplication_roots,
    Duplications = if_else(total != 0, gene_gain_by_duplication / total, 0),
    Losses = if_else(total != 0, gene_loss / total, 0),
    Gains = if_else(total != 0, Gainss / total, 0)
  ) %>%
  filter(!is.na(Gains), !is.na(Losses), !is.na(Duplications), !is.na(node_id))

####### Fathers and sons
edges <- tree@phylo$edge  # Parent-child relationship matrix (parent, child)
colnames(edges) <- c("Parent", "Child")

# Convert edges to a data frame
edges_df <- as.data.frame(edges)

# Join "edges_df" with "tree_data" to retrieve "node_id" corresponding to "node"
edges_df <- edges_df %>%
  left_join(tree_data %>% select(node, node_id), by = c("Parent" = "node")) %>%
  mutate(Parent = ifelse(!is.na(node_id), node_id, Parent)) %>%
  select(-node_id)  # Remove 'node_id' column after replacement

# Join "edges_df" with "tree_data" to retrieve "node_id" corresponding to "Child"
edges_df <- edges_df %>%
  left_join(tree_data %>% select(node, node_id), by = c("Child" = "node")) %>%
  mutate(Child = ifelse(!is.na(node_id), node_id, Child)) %>%
  select(-node_id)  # Remove "node_id" column after replacement

# Create a data frame to associate nodes with parent information
father_info <- tree@data %>%
  mutate(Child = node_id) %>%  # Rename node to Child to match with edges
  select(Child, gene_gain_by_duplication)  # Select node and relevant information

# Combine parent information with the original tree
tree@data <- tree@data %>%
  left_join(edges_df, by = c("node_id" = "Child")) %>%  # Add 'Parent' column
  left_join(father_info, by = c("Parent" = "Child"), suffix = c("", "_father")) %>%  # Add parent information
  rename(father_node = gene_gain_by_duplication_father)

# Recursive function to get all ancestors (parents) of a node
get_all_parents <- function(node, edges_df) {
  # Find the immediate parent
  parents <- edges_df %>%
    filter(Child == node) %>%
    pull(Parent)

  # If no parents, return an empty list
  if (length(parents) == 0) {
    return(NULL)
  }

  # If parents exist, recursively find all ancestors
  all_parents <- parents
  for (parent in parents) {
    # Recursive call to find parents of each parent
    all_parents <- c(all_parents, get_all_parents(parent, edges_df))
  }

  # Return all unique ancestors
  return(unique(all_parents))
}

# Apply the function to retrieve all ancestors (parents)
tree_data_with_all_parents <- tree_data %>%
  rowwise() %>%
  mutate(fathers = list(get_all_parents(node_id, edges_df)))

####### Calculate de novo and ancestral

# Create a new 'total' column by summing values for each row
tree_data_with_all_parents$total <- sapply(seq_along(tree_data_with_all_parents$fathers), function(i) {
  # Extract the current element from "fathers"
  fathers_list <- tree_data_with_all_parents$fathers[[i]]
  own_value <- tree_data_with_all_parents$gene_gain_by_duplication[i]

  # Check if 'fathers_list' is NULL or empty
  if (is.null(fathers_list) || length(fathers_list) == 0) {
    return(own_value)  # If no parents, return own value
  }

  # Convert the list of parents to numeric
  temp_fathers <- as.numeric(fathers_list)

  # Match with node_id and sum "gene_gain_by_duplication"
  temp_sum <- sum(
    tree_data_with_all_parents$gene_gain_by_duplication[
      tree_data_with_all_parents$node_id %in% temp_fathers
    ],
    na.rm = TRUE
  )
  return(temp_sum + own_value)  # Return the sum of parents and own value
})

# Add new columns "de_novo" and "ancestral"
tree_data_with_all_parents <- tree_data_with_all_parents %>%
  mutate(
    Novo = (gene_gain_by_duplication / total) * 100,  # Calculate "de_novo"
    Ancestral = 100 - Novo                           # Calculate "ancestral"
  )

# Prepare data for heatmap
heatmap_data_full <- tree_data_with_all_parents %>%
  select(node, Gains, Losses, Duplications) %>%
  pivot_longer(cols = c("Gains", "Losses", "Duplications"), names_to = "variable", values_to = "value") %>%
  group_by(variable) %>%
  mutate(
    min_value = min(value, na.rm = TRUE),
    max_value = max(value, na.rm = TRUE),
    normalized_value = (value - min_value) / (max_value - min_value)
  ) %>%
  ungroup()

#################################################################################
'''
modules['Module_6']= '''#####################################
# Module 6 | Legends and save plots #
#####################################

# This script will be called by the python program revolutionhtl.plot_summary
# The variables bellow are suposed to be created by python, to reset uncoment lines below
#o_format <- "pdf"
#prefix <- "tl_project."

oname <- paste(prefix, "change_gene_content.", "pdf",sep="")

# Prepare data
heatmap_legend_data <- data.frame(
  variable = c("Gains", "Losses", "Duplications"),
  log2_deviation = seq(log_min, log_max, length.out = 3)
)

heatmap_legend_plot <- ggplot(heatmap_legend_data, aes(x = variable, y = 1, fill = 1)) +
  geom_tile(color = "black") +
  scale_fill_gradientn(
    colors = c("#FF4545", "#FCF596", "#B1D690"),  # Custom colors
    values = scales::rescale(c(0, 0.5, 1), to = c(0, 1)),  # Scale values from 0 to 1
    limits = c(log_min, log_max),  # Keep the range according to log_min and log_max
    name = "Normalized Value",  # Legend label
    na.value = "white", # For NA values
    guide = guide_colorbar(
      title.position = "top",  # Position title at the top
      title.hjust = 0.5        # Center title horizontally
    )
  ) +
  theme_minimal() +
  theme(
    axis.title.x = element_blank(),
    axis.text.x = element_text(angle = 90, hjust = 1, size = 10),
    axis.title.y = element_blank(),
    axis.text.y = element_blank(),
    panel.grid = element_blank(),
    legend.position = "bottom",  # Legend at the bottom
    legend.direction = "horizontal"  # Horizontal legend direction
  )

heatmap_legend_plot

# Legend for the stacked bar plot with variable names
bar_legend_data <- data.frame(
  variable = factor(c("Novo", "Ancestral"), levels = c("Novo", "Ancestral")),
  value = c(0, 0)
)

bar_legend_plot <- ggplot(bar_legend_data, aes(x = 1, y = value, fill = variable)) +
  geom_bar(stat = "identity") +
  scale_fill_manual(
    values = c("Novo" = "#66C2A5", "Ancestral" = "#FC8D62"),
    name = "Duplications"
  ) +
  theme_minimal() +
  theme(
    axis.title.x = element_blank(),
    axis.text.x = element_blank(),
    axis.title.y = element_blank(),
    axis.text.y = element_blank(),
    panel.grid = element_blank()
    #    legend.position = "right", legend.direction = "horizontal"
  )

bar_legend_plot

# Combine the final plot with adjusted legends
if (num_species > 20) {
  # If num_species is greater than 20, exclude bar_legend_plot
  blank_legend <- cowplot::ggdraw() + cowplot::draw_label("", size = 0)  # Blank space

  combined_with_legends <- cowplot::plot_grid(
    final_plot,
    cowplot::plot_grid(heatmap_legend_plot, blank_legend, ncol = 1, rel_heights = c(0.1, 0.3)),  # Include blank instead of bar_legend_plot
    ncol = 2,
    rel_widths = c(9, 1)  # Adjust width of the legend column
  )
} else {
  # If num_species is 20 or less, include bar_legend_plot
  combined_with_legends <- cowplot::plot_grid(
    final_plot,
    cowplot::plot_grid(heatmap_legend_plot, bar_legend_plot, ncol = 1, rel_heights = c(0.1, 0.3)),  # Include both legends
    ncol = 2,
    rel_widths = c(9, 1)  # Adjust width of the legend column
  )
}
# Display the combined plot with legends
# print(final_plot)
print(combined_with_legends)

# Save
ggsave(oname, plot = combined_with_legends, device = "pdf", width = 16, height = 10)
print(paste('Written to ', oname))

#################################################################################
'''
modules['Module_4']= '''### Module 4
### Calculations and prepare data
# Prepare heatmaps

# Calculate log2_deviation_min and log2_deviation_max
log_min <- min(heatmap_data_full$normalized_value, na.rm = TRUE)
log_max <- max(heatmap_data_full$normalized_value, na.rm = TRUE)

colors <- c("#FF4545", "#FCF596", "#B1D690")

# Determine Max value
max_data_value <- tree_data %>%
  select(gene_gain_by_duplication, gene_loss, duplication_roots) %>%
  summarise(across(everything(), max)) %>%
  unlist() %>%
  max()


if (nchar(as.character(max_data_value)) > 4 || num_species > 13) {
  wsize <- 2.5
} else {
  wsize <- 3.5
}


# Custom function to create combined plots for each node
custom_node_combined_plot <- function(data, heatmap_cols, bar_cols, log_min, log_max, num_species, color_palette = NULL) {

  heatmap_data <<- data %>%
    select(node, all_of(heatmap_cols), duplication_roots, gene_gain_by_duplication, gene_loss) %>%
    pivot_longer(cols = all_of(heatmap_cols), names_to = "variable", values_to = "value") %>%
    group_by(variable) %>%
    mutate(
      min_value = min(value, na.rm = TRUE),
      max_value = max(value, na.rm = TRUE),
      normalized_value = (value - min_value) / (max_value - min_value)
    ) %>%
    ungroup()

  # Create count data in long format for the heatmap
  counts_data <<- data %>%
    select(node,
           Gains = Gainss,
           Duplications = gene_gain_by_duplication,
           Losses = gene_loss) %>%
    pivot_longer(cols = c("Gains", "Duplications", "Losses"),
                 names_to = 'variable', values_to = 'count')

  # Merge counts with heatmap data
  heatmap_data <<- heatmap_data %>%
    left_join(counts_data, by = c('node', 'variable'))

  # Prepare data for stacked bar plot
  bar_data <<- data %>%
    select(node, all_of(bar_cols)) %>%
    pivot_longer(cols = all_of(bar_cols), names_to = "variable", values_to = "value")

  # Create a list to store the combined plots
  combined_list <<- list()

  # Define the color palette for the heatmap
  if (is.null(color_palette)) {
    color_palette <- rev(RColorBrewer::brewer.pal(11, "RdBu"))
  }

  # Iterate over each node to create the combined plot
  unique_nodes <<- unique(heatmap_data$node)
  for (node_id in unique_nodes) {
    # Heatmap data for the current node
    node_heatmap_data <<- heatmap_data %>% filter(node == node_id)
    # Bar plot data for the current node
    node_bar_data <- bar_data %>% filter(node == node_id)

    # Calculate the total percentage for the current node (should be 1)
    total_value <- sum(node_bar_data$value, na.rm = TRUE)

    # Add column to identify the segment with the highest value
    node_bar_data <- node_bar_data %>%
      mutate(
        percentage = (value / total_value) * 100,  # Convert to percentage
        is_max = value == max(value, na.rm = TRUE)
      )

    if (num_species > 16) {
      p_heatmap <- ggplot(node_heatmap_data, aes(x = variable, y = 1, fill = normalized_value)) +
        geom_tile(color = "black",
                  width = 1,  # Adjust tile size
                  height = 0.6) +  # Adjust tile height
        geom_text(aes(label = ""),  # Do not show numbers if num_species > 20
                  size = wsize, fontface = "bold", color = "black") +
        scale_fill_gradientn(
          colors = colors,  # Custom color palette
          values = scales::rescale(c(0, 0.5, 1), to = c(0, 1)),  # Scale values from 0 to 1
          limits = c(0, 1),  # Limit values between 0 and 1
          na.value = "white",  # Assign NA values to white to avoid gray
          name = "Normalized Value"
        ) +
        theme_void() +
        theme(legend.position = "none",
              axis.text.y = element_blank(),
              axis.text.x = element_blank(),
              plot.margin = margin(0, 0, 0, 0)) +  # Remove additional margins
        coord_fixed(ratio = 1)  # Maintain square aspect ratio

    } else {
      # If num_species <= 20, show numbers
      p_heatmap <- ggplot(node_heatmap_data, aes(x = variable, y = 1, fill = normalized_value)) +
        geom_tile(color = "black",
                  width = 1,  # Adjust tile size
                  height = 0.6) +  # Adjust tile height
        geom_text(aes(label = count),  # Show numbers if num_species <= 20
                  size = wsize, fontface = "bold", color = "black") +
        scale_fill_gradientn(
          colors = colors,  # Custom color palette
          values = scales::rescale(c(0, 0.5, 1), to = c(0, 1)),  # Scale values from 0 to 1
          limits = c(0, 1),  # Limit values between 0 and 1
          na.value = "white",  # Assign NA values to white to avoid gray
          name = "Normalized Value"
        ) +
        theme_void() +
        theme(legend.position = "none",
              axis.text.y = element_blank(),
              axis.text.x = element_blank(),
              plot.margin = margin(0, 0, 0, 0)) +  # Remove additional margins
        coord_fixed(ratio = 1)  # Maintain square aspect ratio
    }


    p_bar <- NULL
    if (num_species <= 25) {
      p_bar <- ggplot(node_bar_data, aes(x = 1, y = value, fill = variable)) +
        geom_bar(stat = "identity", width = 1) +
        geom_text(
          data = node_bar_data %>% filter(is_max),
          aes(label = paste0(round(percentage, 0), "%"), y = value / 2),
          color = "black", size = 3.5, fontface = "bold"
        ) +
        scale_fill_manual(values = c("Novo" = "#66C2A5", "Ancestral" = "#FC8D62"), name = "Duplications") +
        theme_void() +
        coord_polar("y") +  # Use polar coordinates for circular plot
        theme(legend.position = "none", plot.margin = margin(0, 0, 0, 0))  # Remove margins in circular plot
    }



    # Combine the heatmap and bar plot (if exists)
    if (!is.null(p_bar) & num_species < 16) {
      combined_plot <- grid.arrange(p_heatmap, p_bar, ncol = 1, heights = c(1, 1))
    } else if (!is.null(p_bar) & num_species >= 16) {
      combined_plot <- grid.arrange(p_heatmap, p_bar, nrow = 1, widths = c(1.7, 1))
    } else {
      combined_plot <- p_heatmap
    }

    # Add the combined plot to the list
    combined_list[[as.character(node_id)]] <- combined_plot
  }

  return(combined_list)
}

# Create the combined plots for each node, passing the num_species variable
combined_plots <- custom_node_combined_plot(
  data = tree_data_with_all_parents,
  heatmap_cols = c("Gains", "Losses", "Duplications"),
  bar_cols = c("Novo", "Ancestral"),
  log_min = log_min,
  log_max = log_max,
  num_species = num_species  # Use the num_species variable
)

# Create the final plot with the combined plots inserted
# Define a list of parameters for each range of species
params_list <- list(
  list(min_species = 101, width = 0.028, height = 0.028, hjust = 0.15, vjust = -0.4),
  list(min_species = 91,  width = 0.030, height = 0.030, hjust = 0.15, vjust = -0.4),
  list(min_species = 81,  width = 0.032, height = 0.032, hjust = 0.15, vjust = -0.4),
  list(min_species = 71,  width = 0.035, height = 0.035, hjust = 0.15, vjust = -0.4),
  list(min_species = 61,  width = 0.04,  height = 0.04,  hjust = 0.15, vjust = -0.4),
  list(min_species = 41,  width = 0.05,  height = 0.05,  hjust = 0.2,  vjust = -0.35),
  list(min_species = 31,  width = 0.06,  height = 0.06,  hjust = 0.2,  vjust = -0.3),
  list(min_species = 26,  width = 0.08,  height = 0.08,  hjust = 0.2,  vjust = -0.3),
  list(min_species = 16,  width = 0.1,   height = 0.1,   hjust = 0.12, vjust = -0.25),
  list(min_species = 14,  width = 0.08,  height = 0.08,  hjust = 0.07, vjust = -0.02),
  list(min_species = 11,  width = 0.1,   height = 0.1,   hjust = 0.07, vjust = -0.02),
  list(min_species = 6,   width = 0.12,  height = 0.12,  hjust = 0.07, vjust = -0.05),
  list(min_species = 0,   width = 0.15,  height = 0.15,  hjust = 0.02, vjust = -0.01)
)

# Find the parameters that match the current number of species
selected_params <- params_list[[which.max(sapply(params_list, function(x) num_species >= x$min_species))]]

# Use the selected parameters to create the final plot
final_plot <- inset(p, combined_plots,
                    width = selected_params$width,
                    height = selected_params$height,
                    hjust = selected_params$hjust,
                    vjust = selected_params$vjust)

#################################################################################
'''
modules['Module_5']= '''### Module 5
### Bar plot
# Prepare data

df_change_new <- p[["data"]] %>%
  filter(!is.na(label)) %>%
  mutate(Change_gene_content = duplication_roots + gene_gain_by_duplication - gene_loss) %>%
  select(label, Change_gene_content) %>%
  gather(key = "category", value = "value", -label)

# Get non-NA values for label
label_order <- unique(na.omit(p[["data"]]$label))
label_order <- rev(label_order)  # Reverse the order

# Now use this order in the 'label' column
df_change_new$label <- factor(df_change_new$label, levels = label_order)

#
if (num_species >= 2 && num_species <= 5) {
  bar_width <- 0.2
} else if (num_species > 5 && num_species <= 10) {
  bar_width <- 0.3
} else {
  bar_width <- 0.5
}
#

# Plot the result

p_bar1 <- ggplot(df_change_new, aes(x = value, y = category)) +
  geom_bar(
    aes(fill = value > 0),  # Conditional coloring based on whether the value is positive or negative
    stat = "identity",
    width = bar_width
  ) +
  scale_fill_manual(
    values = c("#B03052", "#4f93fcff"),  # Dark red for negative values, blue for positive ones
    name = "Sign"
  ) +
  # Vertical line at x = 0
  geom_vline(
    xintercept = 0,
    color = "black",  # Black line
    linewidth = 1     # Line thickness
  ) +
  facet_wrap(~ label, ncol = 1) +
  theme_minimal() +
  theme(
    legend.position = "none",               # Remove legend
    panel.grid = element_blank(),           # Remove grid
    strip.text = element_blank(),           # Remove facet titles
    axis.text.y = element_blank(),          # Remove Y-axis text
    axis.title.y = element_blank(),         # Remove Y-axis title
    axis.title.x = element_blank(),  # Adjust X-axis title size
    plot.title = element_text(size = 15, hjust = 0.5)
  ) +
  ggtitle("Change gene content")

# Display the bar plot to check
print(p_bar1)

# Combine the tree plot with the bar plot
final_plot <- plot_grid(final_plot, p_bar1, rel_widths = c(5, 1))

print(final_plot)

#################################################################################
'''
modules= Series(modules)



"""
Functions to run blast

Includes standalon usage:
python -m revolutionhtl.blast -h
"""

from .common_tools import norm_path
from .error import DirectoryExist

from tqdm import tqdm
from itertools import combinations
import os
from subprocess import run
from joblib import Parallel, delayed
# constants
_default_out_dir= 'tl_project_alignment_all_vs_all'
_default_fasta_ext='.fa'
_default_blast='blastn'
_default_makeblastdb='makeblastdb'
_blast_columns= 'qseqid sseqid qlen slen length bitscore evalue'

# Functions
def blast_all_vs_all(fastas_dir,
                     out_dir= _default_out_dir,
                     fasta_ext= _default_fasta_ext,
                     blast_command= _default_blast,
                     makeblastdb= _default_makeblastdb,
                     evalue= 0.00001,
                     identity= 25,
                     cov= 50,
                     max_target_seqs= 25,
                     num_threads=1,
                     num_jobs=1
                    ):
    # Normalize and check directories
    fastas_dir, out_dir= _check_path(fastas_dir, out_dir)
    # List fasta files
    is_fasta= lambda x: x.endswith(fasta_ext)
    fasta_files= list(filter(is_fasta, os.listdir(fastas_dir)))
    # Create file names
    species= {file: '.'.join(file.split('.')[0:-1])
              for file in fasta_files}
    reference= {file:f'{out_dir}{species[file]}' for file in fasta_files}
    f_path= {file:f'{fastas_dir}{file}' for file in fasta_files}
    # Set fasta type
    dbtype= _blast_2_seq_type(blast_command)
    # Create index
    for file in tqdm(fasta_files, desc= 'Generating index'):
        std_out= blast_makedb(makeblastdb, f_path[file], reference[file], dbtype)
    # Run pairwise
    n= len(fasta_files)
    if num_jobs==1:
        for query ,target in tqdm(combinations(fasta_files, 2), total= int(n*(n-1)/2),
                      desc= 'Running alignments'):

            fout= f'{out_dir}{species[query]}.vs.{species[target]}.blast.alignment_hits'
            std_out= blast_blastp(blast_command, f_path[query],
                                reference[target], fout,
                                evalue= evalue, identity= identity, cov= cov,
                                max_target_seqs= max_target_seqs, num_threads=num_threads
                               )

            fout= f'{out_dir}{species[target]}.vs.{species[query]}.blast.alignment_hits'
            std_out= blast_blastp(blast_command, f_path[target],
                                reference[query], fout,
                                evalue= evalue, identity= identity, cov= cov,
                                max_target_seqs= max_target_seqs, num_threads=num_threads
                               )
    else:
        blast_both_ways = lambda query, target: ( blast_blastp(blast_command, f_path[query],
                                reference[target], f'{out_dir}{species[query]}.vs.{species[target]}.blast.alignment_hits',
                                evalue= evalue, identity= identity, cov= cov,
                                max_target_seqs= max_target_seqs, num_threads=num_threads
                               ), blast_blastp(blast_command, f_path[target],
                                reference[query], f'{out_dir}{species[target]}.vs.{species[query]}.blast.alignment_hits',
                                evalue= evalue, identity= identity, cov= cov,
                                max_target_seqs= max_target_seqs, num_threads=num_threads
                               ))
        Parallel(n_jobs=num_jobs,verbose=3)(delayed(blast_both_ways)(query,target) for query,target in combinations(fasta_files, 2))

    # End

def blast_makedb(makeblastdb, fpath, f_reference, dbtype):
    command= [makeblastdb, '-in', fpath,
              '-dbtype', dbtype,
              '-out', f_reference,
             ]
    return run(command)

def blast_blastp(blast_command, fpath, f_reference, f_out, evalue= 0.00001, identity= 25, cov= 50, max_target_seqs= 25, num_threads=1):
    command= [blast_command,
              '-num_threads', num_threads,
              '-db', f_reference, '-query', fpath, '-out', f_out,
              '-evalue', evalue,
              '-qcov_hsp_perc', cov,
              '-max_target_seqs', max_target_seqs,
              '-outfmt', f'6 {_blast_columns}'
             ]
    if blast_command=='blastn':
        command+= ['-perc_identity', identity]

    command= list(map(str, command))
    return run(command)

# Auxiliary function
####################

def _check_path(fastas_dir, out_dir):
    # Fastas directory
    fastas_dir= norm_path(fastas_dir)
    if not os.path.exists(fastas_dir):
        raise DirectoryExist(f'There is not directory "{fastas_dir}"')
    # ourput directory
    if out_dir==None:
        out_dir= fastas_dir
    else:
        out_dir= norm_path(out_dir)
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
    # return
    return fastas_dir, out_dir

def _blast_2_seq_type(blast_command):
    if 'blastn' in blast_command:
        return 'nucl'
    elif 'blastp' in blast_command:
        return 'prot'
    elif 'blastx' in blast_command:
        raise ValueError('blastx not suported yet')
    raise ValueError(f'Command {blast_command} not recognized')
    return None

# Stand-alone usage
###################

if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('fastas_dir', type= str)
    parser.add_argument('-o', '--out_dir', default= _default_out_dir, type= str)
    parser.add_argument('-f_ext', '--fasta_ext', default= _default_fasta_ext, type= str)
    parser.add_argument('-b_command', '--blast', default= _default_blast, type= str, choices= ['blastn', 'blastp'])
    parser.add_argument('-m_command', '--makeblastdb', default= _default_makeblastdb, type= str)
    parser.add_argument('-e', '--evalue', default= 0.00001, type= float)
    parser.add_argument('-id', '--identity', default= 25, type= int)
    parser.add_argument('-cov', '--coverture', default= 50, type= int)
    parser.add_argument('-k', '--max_target_seqs', default= 25, type= int)
    args = parser.parse_args()

    # Run
    ######
    hello_blast= "\n-------------------------------------"
    hello_blast+= "\nRunnng BLAST to compute alignment hits"
    hello_blast+= "\n------------------------------------\n"
    print(hello_blast)
    blast_all_vs_all(args.fastas_dir,
                     out_dir= args.out_dir,
                     fasta_ext= args.fasta_ext,
                     blast_command= args.blast,
                     makeblastdb= args.makeblastdb,
                     evalue= args.evalue,
                     identity= args.identity,
                     cov= args.coverture,
                     max_target_seqs= args.max_target_seqs,
                    )
    print("Process finished succesffully\n")



def norm_path(path):
    if not path.endswith('/'):
        return path + '/'
    return path

def txtMu_2_dict(txt):
    return dict(map(lambda x: x.split(':'), txt.split(',')))

def update_mu(T0, T1, mu, attr0, attr1):
    DD0= attr_2_node(T0, attr0)
    DD1= attr_2_node(T1, attr1)
    return {DD0[x0]:DD1[x1] for x0,x1 in mu.items()}

def attr_2_node(T, attr):
    ret= {T.nodes[node][attr] : node for node in T}
    if len(ret)!=len(T):
        raise ValueError('There are nodes with the same attribute.')
    return ret

Devents= {'speciation' : 'S', 'duplication' : 'D', 'S':'S','D':'D','X':'X'}
def add_eventAttr(Tg, oAttr, nAttr):
    for xNode in Tg:
        ev= Tg.nodes[xNode][oAttr]
        Tg.nodes[xNode][nAttr]= Devents.get(ev, 'S')
    return Tg



desc_def= """ | Steps :
1. Alignment hits computation
2. Best hit selection
3. Gene tree reconstruction and orthology inference
4. Species tree reconstruction
5. Gene and species trees reconciliation |
More details at: https://pypi.org/project/revolutionhtl/
"""

if __name__ == "__main__":
    from importlib.metadata import version
    from .constants import _default_f
    from .common_tools import norm_path
    import argparse

    V_tl= version('revolutionhtl')
    V_tl

    txt= f'REvolutionH-tl: Reconstruction of Evolutionaty Histories TooL (V{V_tl})'

    parser = argparse.ArgumentParser(prog= 'revolutionhtl',
                                     description=f'{txt}{desc_def}',
                                     usage='python -m revolutionhtl <arguments>',
                                     formatter_class=argparse.MetavarTypeHelpFormatter,
                                    )

    # Arguments
    ###########

    # Input data
    # ..........
    parser.add_argument('-steps',
                        help= 'List of steps to run (default: 1 2 3 4).',
                        type= int,
                        nargs= '*',
                        default= [1, 2, 3, 4, 5]
                       )

    parser.add_argument('-F', '--fastas',
                        help= '[str | Input for step 1] Directory containing fasta files.',
                        type= str,
                        default= None
                       )

    parser.add_argument('-alignment_h', '--alignment_hits',
                        help= '[str | Input for step 2] Directory containing alignment hits.',
                        type= norm_path,
                        default= None
                       )

    parser.add_argument('-best_h', '--best_hits',
                        help= '[str | Input for step 3] .tsv file containing best hits.',
                        type= str,
                        default= None,
                       )

    parser.add_argument('-T', '--gene_trees',
                        help= '[str | Input for steps 4 and 5] .tsv file containing gene trees in nhx format.',
                        type= str,
                       )

    parser.add_argument('-S', '--species_tree',
                        help= '[str | Input for step 5] .nhx file containing a species tree.',
                        type= str,
                        default= None,
                       )

    # Parameters
    # ..........
    al_default= 'diamond'
    parser.add_argument('-aligner', '--aligner',
                        help= f'[str | Parameter for step 1 | Default: {al_default}] Command or path to the program for computation of aligment hits. Supported: diamond, blastn, blastp, blastx',
                        type= str,
                        default= al_default,
                        #choices= ['diamond', 'blastn', 'blastp', 'blastx']
                       )
    parser.add_argument('-v', '--verbose_diamond', action='store_true',
                        help= '[flag | Parameter for step 1 ] Use to display diamond messages.')
    e_default= 0.00001
    parser.add_argument('-e', '--evalue', default= e_default, type= float,
                        help= f'[float | Parameter for step 1 | Default: {e_default}] Maximum evalue required to consider significant an aligment hit.'
                       )
    mk_def= 'makeblastdb'
    parser.add_argument('-m_command', '--makeblastdb', default= mk_def, type= str,
                        help= f'[str | Parameter for step 1 | Default: {mk_def}] BLAST command or path to the BLAST program for database creation.')
    id_def= 25
    parser.add_argument('-id', '--identity', default= id_def, type= int,
                        help= f'[float | Parameter for step 1 | Default: {id_def}] Minimum percentage of identity required to report an alignment hit.')
    cov_def= 50
    parser.add_argument('-cov', '--coverture', default= cov_def, type= int,
                        help= f'[float | Parameter for step 1 | Default: {cov_def}] Minimum percentage of query coverture required to report an alignment hit.')
    kh_def= 100
    parser.add_argument('-k_hits', '--max_target_seqs', default= kh_def, type= int,
                        help= f'[int | Parameter for step 1 | Default: {kh_def}] Maximum number of alignment hits per gene againist a fixed species.')

    num_threads_def= 1
    parser.add_argument('-t', '--num_threads', default= num_threads_def, type= int,
                        help= f'[int | Parameter for step 1 | Default: {num_threads_def}] Number of threads to use for diamond or blastp.')

    num_jobs_def= 1
    parser.add_argument('-j', '--num_jobs', default= num_jobs_def, type= int,
            help= f'[int | Parameter for step 1 | Default: {num_jobs_def}] Number of concurrent jobs to run for the aligner. Note: the total number of threads used will be num_threads*num_jobs. ')

    bhh_def= 'target'
    from .parse_prt import normalization_modes
    parser.add_argument('-bh_heuristic', '--besthit_heuristic',
                        help= f'[str | Parameter for step 2 | Default: {bhh_def}] Indicates how to normalize bit-score. Normalize by sequence lenght: query, target, alignment, smallest. No normalization: raw.',
                        type= str,
                        choices= normalization_modes,
                        default= bhh_def,
                       )

    parser.add_argument('-f', '--f_value',
                        help= f'[float | Parameter for step 2 | Defualt: {_default_f}] Number between 0 and 1. Defines the adaptative threshhold for best-hit selection as: f*max_bit_score.',
                        type= float,
                        default= _default_f,
                       )

    com_def= 'Louvain'
    com_opt= ['Mincut', 'BPMF', 'Karger', 'Greedy', 'Gradient_Walk', 'Louvain', 'Louvain_Obj']
    parser.add_argument('-bmg_h', '--bmg_heuristic',
                        help= f'[str | Parameter for step 3 | Defult: {com_def}] Comunity detection method for MaxConsistentTriples heuristic. Options: {", ".join(com_opt)}.',
                        type= str,
                        default= com_def,
                        choices= com_opt,
                       )

    parser.add_argument('-no_binary_R', '--no_binary_triples',
                        help= '[flag | Parameter for step 3 ] Use to avoid the usage of binary triples from best-hit graph.',
                        action= 'store_false',
                       )


    parser.add_argument('-f_bT', '--force_binary_gene_tree',
                        help= '[flag | Parameter for step 3 ] Use to force gene trees to be binary.',
                        action= 'store_true',
                       )

    parser.add_argument('-T_no_db', '--gene_tree_no_double_build',
                        help= '[flag | Parameter for step 3 ] Use to avoid running build twice in the MaxConsistentTriples heuristic.',
                        action= 'store_false',
                       )

    com_def= 'louvain_weight'
    com_opt= ['naive', 'louvain', 'mincut', 'louvain_weight']
    parser.add_argument('-stree_h', '--species_tree_heuristic',
                        help= f'[str | Parameter for step 4 | Default: {com_def}] Comunity detection method for MaxConsistentTriples heuristic. Options: {", ".join(com_opt)}.',
                        type= str,
                        default= com_def,
                        choices= com_opt
                       )

    sr_def= 3
    parser.add_argument('-streeh_repeats', '--stree_heuristic_repeats',
                        help= f'[int | Parameter for step 4 | Default: {sr_def}] Specifies how many times run the MaxConsistentTriples heuristic.',
                        type= int,
                        default= sr_def
                       )

    parser.add_argument('-streeh_b', '--streeh_binary',
                        help= '[flag | Parameter for step 4] Use to force specis tree to be binary.',
                        action= 'store_true',
                       )

    parser.add_argument('-streeh_ndb', '--streeh_no_doble_build',
                        help= '[flag | Parameter for step 4] Use to avoid running build twice in the MaxConsistentTriples heuristic.',
                        action= 'store_false',
                       )

    parser.add_argument('-n_edit_T', '--do_not_edit_inconsistent_trees',
                        help= '[flag | Prameter for step 5] Use to avoid editing of inconsistent gene trees.',
                        action= 'store_true',
                       )
    # Format parameters
    # .................

    pr_def= 'tl_project'
    parser.add_argument('-o', '--output_prefix',
                        help= f'[str | Default: {pr_def}] Prefix for output files.',
                        type= str,
                        default= pr_def,
                       )

    fa_def= '.fa'
    parser.add_argument('-fext', '--fasta_ext',
                        help= f'[str | Default: {fa_def}] Extesion for fasta files.',
                        type= str,
                        default= fa_def,
                       )

    parser.add_argument('-singletons',
                        help= "[flag | Prameter for step 2] Use to print singletons, i.e. genes that weren't assigned to an orthogroup.",
                        action= 'store_false',
                       )

    parser.add_argument('-OG', '--orthogroup_file',
                        help= f'[str | Optional parameter for step 2] File specifying orthogroups for the best match graph construction.',
                        type= str,
                        default= None,
                       )

    og_def= 'OG'
    parser.add_argument('-og', '--orthogroup_column',
                        help= f'[str | Default: {og_def}] Column specifying orthogroup ID in input and output .tsv files.',
                        type= str,
                        default= og_def,
                       )

    nm_def= 2000
    parser.add_argument('-Nm', '--N_max',
                        help= f'[int | Default: {nm_def}] Maximum number of genes in a orithogroup, bigger orthogroups are splitted. If 0, no orthogroup is splitted.',
                        type= int,
                        default= nm_def,
                       )

    ks_def= 100
    parser.add_argument('-k', '--k_size_partition',
                        help= f'[int | Default: {ks_def}] Range in the numer of genes of orthogroups procesed in batch: first those with <k genes, then <2k. then <3k, and so on.',
                        type= int,
                        default= ks_def,
                       )

    sep_def= ';'
    parser.add_argument('-S_attr', '--S_attr_sep',
                        help= f'[str | Default: {sep_def}] Attribute delimiter in the input .nhx file (Input of step 5).',
                        type= str,
                        default= sep_def,
                       )

    args= parser.parse_args()

    ################
    # Process data #
    ################
    from .error import MissedData, ParameterError
    from .nhx_tools import read_nhx, get_nhx
    from pandas import DataFrame, Series, read_csv
    from os import listdir, cpu_count
    from datetime import datetime

    from .hello import hello5
    print(f'{hello5}V{V_tl}\n')

    allowed_steps= {1, 2, 3, 4, 5}
    bad_steps= set(args.steps) - allowed_steps
    if len(bad_steps) > 0:
        raise ParameterError(f'Only steps 1, 2, 3, 4, and 5 are allowed to be used in the parameter -steps. ')
    else:
        args.steps= sorted(set(args.steps))

    print(f'Running steps {", ".join(map(str, args.steps))}')

    # set num_threads num_jobs
    cpu_count=cpu_count()
    num_threads=args.num_threads
    if num_threads <= 0:
        num_threads = num_threads_def
    if num_threads > cpu_count:
        num_threads=cpu_count
    num_jobs=args.num_jobs
    if num_jobs <= 0:
        num_jobs = num_jobs_def
    if num_jobs*num_threads > cpu_count:
        num_jobs=cpu_count//num_threads
    total_workers= num_jobs*num_threads
    print(f'Running {num_jobs} jobs usign {num_threads} threads each.')

    logfile= f'tl_log.txt'
    with open(logfile, 'a') as F:
        DD= vars(args)
        keys= sorted(DD)
        log= (f'{k}={DD[k]}' for k in keys)
        log= '\n'.join(log)+'\n'
        log= f'> {datetime.now()}\n> REvolutionH-tl V{V_tl}\n{log}'
        log= f'----------------------------------------------\n{log}'
        F.write(log)

    try:

        # 1. Alignment hits computation
        ###############################

        if 1 in args.steps:
            if type(args.fastas)!=str:
                raise MissedData('Step 1 needs a value for the parameter --fastas')
            print('\nStep 1: Alignment hits computation')
            print('----------------------------------')
            alignment_hits= f'{args.output_prefix}_alignment_all_vs_all/'

            if 'diamond' in args.aligner:
                from .diamond import diamond_all_vs_all
                diamond_all_vs_all(args.fastas,
                                   out_dir= alignment_hits,
                                   fasta_ext= args.fasta_ext,
                                   diamond= args.aligner,
                                   quiet= not args.verbose_diamond,
                                   evalue= args.evalue,
                                   identity= args.identity,
                                   cov= args.coverture,
                                   max_target_seqs= args.max_target_seqs,
                                   num_threads=num_threads,
                                   num_jobs=num_jobs
                                   )
            elif 'blast' in args.aligner:
                from .blast import blast_all_vs_all
                blast_all_vs_all(args.fastas,
                                 out_dir= alignment_hits,
                                 fasta_ext= args.fasta_ext,
                                 blast_command= args.aligner,
                                 makeblastdb= args.makeblastdb,
                                 evalue= args.evalue,
                                 identity= args.identity,
                                 cov= args.coverture,
                                 max_target_seqs= args.max_target_seqs,
                                 num_threads=num_threads,
                                 num_jobs=num_jobs
                        )
            else:
                ValueError(f'Unrecognized aligner "{args.aligner}"')
            with open(logfile, 'a') as F:
                F.write('> Step 1 completed\n')
            print(f'Alignment hits successfully written to {alignment_hits}')
            if (2 in args.steps) and (type(args.alignment_hits)!=str):
                print('This data will be used as input of step 2')
        else:
            alignment_hits= None

        # 2. Best hit selection
        #######################
        if 2 in args.steps:
            from .parse_prt import _parse
            print("\nStep 2: Best hit selection")
            print("--------------------------")
            generated_hits= type(alignment_hits)==str
            inputed_hits= type(args.alignment_hits)==str
            if (not generated_hits) and (not inputed_hits):
                raise MissedData('Step 2 needs a value for the parameter --alignment_hits. You can create it running step 1')
            elif inputed_hits:
                alignment_hits= norm_path(args.alignment_hits)
            else:
                alignment_hits= norm_path(alignment_hits)
            if args.singletons:
                if type(args.fastas)==str:
                    fasta_path= norm_path(args.fastas)
                    text= f'\nInput fasta files: {fasta_path}'
                    fasta_files= {f'{fasta_path}{x}' for x in listdir(fasta_path) if x.endswith(args.fasta_ext)}
                else:
                    raise MissedData('Step 2 needs a value for the parameter --fastas whenerver you use the -singletons flag')
            else:
                fasta_files= None
                text= ''

            print(f'Input hits: {alignment_hits}{text}')

            if args.orthogroup_file:
                print(f'Input orthogroups: {args.orthogroup_file}')
                from .parse_prt import load_orthogroups_df
                OGs_table= load_orthogroups_df(args.orthogroup_file)
            else:
                OGs_table= None

            df_hits, OGs_table, singletons= _parse(alignment_hits,
                                                    args.f_value,
                                                    args.singletons,
                                                    fasta_files,
                                                    args.besthit_heuristic,
                                                    N_max= args.N_max,
                                                    n_jobs= total_workers,
                                                    OGs_table= OGs_table,
                                                    )

            opath= args.output_prefix+'.best_hits.tsv'
            df_hits.reset_index(drop=True).to_csv(opath, sep='\t', index= False)
            print(f'Best hits successfully written to {opath}')
            if (3 in args.steps) and (type(args.best_hits)!=str):
                print('This file will be used as input for step 3')

            if not args.orthogroup_file:
                from pandas import isna
                opath= args.output_prefix+'.orthogroups.tsv'
                speciesCols= list(OGs_table.columns[3:])
                Fstr= lambda X: X if isna(X) else ','.join(X)
                OGs_table[speciesCols]= OGs_table[speciesCols].map(Fstr)
                OGs_table.to_csv(opath, sep='\t', index= False)
                del OGs_table
                print(f'Orthogroups successfully written to {opath}')

            if args.singletons:
                opath= args.output_prefix+'.singletons.tsv'
                singletons.to_csv(opath, sep='\t', index= False)
                print(f'singletons successfully written to {opath}')
                del singletons

            with open(logfile, 'a') as F:
                F.write('> Step 2 completed\n')
        else:
            df_hits= None

        # 3. Gene tree reconstruction and orthology inference
        ###############################################
        if 3 in args.steps:
            from .hug_free import build_graph_series
            from .orthology import orthologs_from_trees_df
            from .in_out import read_tl_digraph, tl_digraph_from_pandas
            print("\nStep 3: Gene tree reconstruction and orthology inference")
            print("--------------------------------------------------------")
            computedHs= type(df_hits)==DataFrame
            inputedHs= type(args.best_hits)==str
            if (not computedHs) and (not inputedHs):
                raise MissedData('Step 3 needs a value for the parameter -best_hits. You can create it by running step 2')
            if inputedHs:
                print(f'Processing best hit graphs at {args.best_hits}...')
                best_hit_G= read_tl_digraph(args.best_hits, og_col= args.orthogroup_column)
            else:
                print('Processing best hit graphs of step 2...')
                best_hit_G= tl_digraph_from_pandas(df_hits, og_col= args.orthogroup_column)
                del df_hits

            # Store bit score and add weight for missed (non symmetric) hits
            F= lambda x,y,G0: (G0.nodes[x]['accession'], G0.nodes[y]['accession'])
            DD= {F(x,y,G0): G0[x][y]['weight'] for G0 in best_hit_G for (x,y) in G0.edges}
            missed= [(y,x) for (x,y) in DD if (y,x) not in DD]
            DD.update({(y,x): DD[(x,y)] for (y,x) in missed})

            print('Reconstructing gene trees...')
            Tg= build_graph_series(best_hit_G, args)
            del best_hit_G

            if (4 in args.steps) and (type(args.gene_trees)!=str):
                print('These gene trees will be used as input for step 4')
            if (5 in args.steps) and (type(args.gene_trees)!=str):
                print('These gene trees will be used as input for step 5')

            # Compute orthologs
            df_orthologs= orthologs_from_trees_df(Tg.reset_index(), forbidden_leaf_name= 'X', tree_col= 'tree')
            # Add bit-score
            df_orthologs['Normalized_bit_score']= df_orthologs.apply(lambda row: DD.get((row.a, row.b),'*') , axis=1 )
            del DD

            # Print orthologs
            opath= f'{args.output_prefix}.orthologs.tsv'
            df_orthologs.to_csv(opath, sep= '\t', index= False)
            del df_orthologs
            print(f'Orthologs successfully written to {opath}')

            with open(logfile, 'a') as F:
                F.write('> Step 3 completed\n')

        else:
            Tg= None

        # 4. Reconstruct species tree
        #############################
        if 4 in args.steps:
            from .supertree import from_gene_forest
            print("\nStep 4: Species tree reconstruction")
            print("-----------------------------------")
            computedTg= type(Tg)==Series
            inputedTg= type(args.gene_trees)==str

            if (not computedTg) and (not inputedTg):
                raise MissedData('Step 4 needs a value for the parameter -gene_trees. You can create it by running step 3')
            if inputedTg:
                print(f'Reading gene trees from {args.gene_trees}...')
                gTrees= read_csv(args.gene_trees, sep= '\t').set_index(
                    args.orthogroup_column).tree.apply(
                    lambda x: read_nhx(x, name_attr= 'accession'))
            else:
                print(f'Using gene trees generated at step 3...')
                gTrees= Tg

            print("Reconstructing species tree...")
            species_tree= from_gene_forest(gTrees,
                                           method= args.species_tree_heuristic,
                                           numb_repeats= args.stree_heuristic_repeats,
                                           doble_build= args.streeh_no_doble_build,
                                           binary= args.streeh_binary,
                                          )

            s_newick= get_nhx(species_tree, root= 1, name_attr= 'species', ignore_inner_name= True) + '\n'
            opath= f'{args.output_prefix}.species_tree.nhx'
            with open(opath, 'w') as F:
                F.write(s_newick+'\n')
            print(f'Species tree successfully written to {opath}')
            if (5 in args.steps) and (type(args.species_tree)!=str):
                print('This species tree will be used as input for step 5')

            with open(logfile, 'a') as F:
                F.write('> Step 4 completed\n')
        else:
            species_tree= None

        # 5. Reconciliate gene trees and species tree
        ##############################################
        if 5 in args.steps:
            from .reconciliation import reconciliate_many, recon_table_to_str
            from networkx import DiGraph
            print("\nStep 5: Gene and species trees reconciliation")
            print("---------------------------------------------")
            computedTg= type(Tg)==Series
            inputedTg= type(args.gene_trees)==str
            if (not computedTg) and (not inputedTg):
                raise MissedData('Step 5 needs a value for the parameter -gene_trees. You can create it by running step 3')
            if inputedTg:
                print(f'Reading gene trees from {args.gene_trees}...')
                gTrees= read_csv(args.gene_trees, sep= '\t').set_index(
                    args.orthogroup_column).tree.apply(
                    lambda x: read_nhx(x, name_attr= 'accession'))
            else:
                print(f'Using gene trees generated at step 3...')
                gTrees= Tg


            computedTs= type(species_tree)==DiGraph
            inputedTs= type(args.species_tree)==str
            if (not computedTs) and (not inputedTs):
                raise MissedData('Step 5 needs a value for the parameter -species_tree. You can create it by running step 4')
            if inputedTs:
                print(f'Reading species tree from {args.species_tree}...')
                with open(args.species_tree) as F:
                    sTree= read_nhx(''.join( F.read().strip().split('\n') ),
                                     name_attr= 'species', attr_sep= args.S_attr_sep)
            else:
                print(f'Using species tree generated at step 4...')
                sTree= species_tree

            if not args.do_not_edit_inconsistent_trees:
                from .tree_correction import correct_tree_df
                gTrees= gTrees.reset_index().set_index(args.orthogroup_column)
                print('Editing inconsistent gene trees...')
                gTrees= correct_tree_df(gTrees, sTree, tree_col= 'tree', root= 1,
                                        label_attr= 'accession',
                                        species_attr= 'species',
                                        event_attr= 'accession',
                                        inplace= False
                                       )
                F= lambda T: get_nhx(T, root= T.root, name_attr= 'accession')
                for_file= gTrees.reset_index()
                for_file.tree= for_file.tree.apply(F)

                opath= f'{args.output_prefix}.corrected_trees.tsv'
                for_file.to_csv(opath, sep= '\t', index= False)
                print(f'Corrected gene trees successfully written to {opath}')
                gTrees= gTrees.tree
                root= None
            else:
                root= 1

            print('Reconciling trees...')
            df_recs= reconciliate_many(gTrees, sTree, root= root)

            # Write resolved trees
            df_r= recon_table_to_str(sTree, df_recs, args.orthogroup_column)
            opath= f'{args.output_prefix}.reconciliation.tsv'
            df_r.to_csv(opath, sep= '\t', index= False)
            print(f'Reconciliation successfully written to {opath}')

            # Write labeled species tree
            nhx_s= get_nhx(sTree, name_attr= 'species', root= 1, ignore_inner_name= True)
            opath= f'{args.output_prefix}.labeled_species_tree.nhx'
            with open(opath, 'w') as F:
                F.write(nhx_s+'\n')
            print(f'Indexed species tree successfully written to {opath}')

            with open(logfile, 'a') as F:
                F.write('> Step 5 completed\n')
        print("\nREvolutionH-tl finished all the tasks without any problem")
        print("---------------------------------------------------------\n")

    except MissedData as e:
        eMessage= f"{e}"
        message= f"\nERROR:\n{eMessage}\n"
        message+= "Please provide proper input data.\n"
        message+= "See the documentation at https://pypi.org/project/revolutionhtl/\n"
        message+= "or run 'python -m revolutionhtl -h' to display help message.\n"
        print(message)

        with open(logfile, 'a') as F:
            F.write(eMessage+'\n'+"---------------------------------------------------------\n")



import pandas as pd
import os
import numpy as np
import networkx as nx
from itertools import chain
from tqdm import tqdm
tqdm.pandas()
from Bio.SeqIO.FastaIO import SimpleFastaParser
from joblib import Parallel, delayed

from .common_tools import norm_path
from .in_out import create_cBMG, tl_digraph_from_pandas
from .reduce_giant import reduce_graph
from .error import MissedData, ParameterError

####################
#                  #
# Some parameters  #
#                  #
####################

from .constants import _diamond_tl_headers, _default_f, _df_matches_cols

####################
#                  #
# Parse functions  #
#                  #
####################

normalization_modes= ['query',
                      'target',
                      'alignment',
                      'smallest',
                      'biggest',
                      'raw',
                      'Pdist',
                      'Pdist-JC',
                      'Pdist-LG',
                      'Pdist-Dayoff',
                     ]

from tqdm_joblib import tqdm_joblib

def _parse(path, f_value, singletons, fastaspath, mode= 'normal', N_max= 2000, n_jobs= 1, OGs_table= None):
    if mode not in normalization_modes:
        raise ParameterError(f'Mode for parsing only can be one of: {", ".join(normalization_modes)}')
    # Load best hits
    df_best= load_best_hits(path, mode, f_value, n_jobs, OGs_table= OGs_table)
    if OGs_table is None:
        # Identify orthogroups
        df_best, OGs_table= identify_orthogroups(df_best, N_max= N_max)
    if singletons:
        singletons= identify_singletons(OGs_table, fastaspath)
    return df_best, OGs_table, singletons

def load_orthogroups_df(path, OG_col= 'OG'):
    df= pd.read_csv(path, sep='\t')
    species= df.columns[3:]
    df[species]= df[species].map(lambda x: x if pd.isna(x) else set(x.split(',')))
    return df

def get_gene_2_orthogroup(df_OGs):
    # Dictionary orthogroup->genes
    species= df_OGs.columns[3:]
    OG_2_genes= df_OGs[species].apply(lambda X: set(chain.from_iterable(X.dropna())), axis= 1)
    # Dictionary gene -> orthogroup
    gene_2_OG= {j:i for i in OG_2_genes.index for j in OG_2_genes.loc[i]}
    N= sum((1 for i in OG_2_genes.index for j in OG_2_genes.loc[i]))
    if len(gene_2_OG) != N:
        raise ValueError('There are repeated genes')
    return gene_2_OG

def load_best_hits(path, mode, f_value, n_jobs, OGs_table= None):
    species_pairs= identify_file_pairs(path)
    # Select best hits for each pair of species
    if OGs_table is None:
        gene_2_OG= None
    else:
        gene_2_OG= get_gene_2_orthogroup(OGs_table)

    F= lambda X: pair_best_hits(X, path, mode, f_value, gene_2_OG=gene_2_OG)
    if n_jobs == 1:
        df_best= map(F, species_pairs.items())
    else:
        with tqdm_joblib(tqdm(desc="Selecting best hits", total=len(species_pairs))) as progress_bar:
            df_best= Parallel(n_jobs=n_jobs)(delayed(F)(X) for X in species_pairs.items())
    # Combine results for every pair of species
    df_best= pd.concat(df_best)
    # Check that files are not empy

    collected_species= set(map(frozenset, df_best[['Query_species','Target_species']].values))
    expected_species= set(map(frozenset, species_pairs))
    missing_species= expected_species - collected_species

    if len(missing_species)>0:
        txt= ','.join(map(lambda X: '-vs-'.join(X), missing_species))
        txt= f'There are no hits for: {txt}'
        raise MissedData(txt)

    if df_best.shape[0]==0:
        raise MissedData('There are not alignment hits in the input directory.')

    return df_best

def load_all_hits_raw(path):
    species_pairs= identify_file_pairs(path)
    # Load raw hits for each pair of species
    F= lambda X: pair_all_hits_raw(X, path)
    df_hits= map(F, species_pairs.items())
    return pd.concat(df_hits)


ext_alignment_hits= '.alignment_hits'
def identify_file_pairs(path):
    alignment_files= filter(lambda file: file.endswith(ext_alignment_hits), os.listdir(path))
    species_pairs= dict()
    for file in alignment_files:
        species_AB= simple_species_parser(file)
        species_BA= species_AB[::-1]
        if species_BA in species_pairs:
            species_pairs[species_BA]+= [file]
        else:
            species_pairs[species_AB]= [file]
    missing_files= [x for x,y in species_pairs.items() if len(y)<2]
    if len(missing_files)>0:
        TEXT= f'There are missing bidirectional-alignment hit files for the pairs of species:\n{ missing_files }'
        raise MissedData(TEXT)
    return species_pairs

def pair_best_hits(X, path, mode, f_value, gene_2_OG=None):
    species_list,file_list= X
    # Load alignment hits
    F1= [path+x for x in file_list]
    df_hits= parse_prt_hits(F1, species_list)
    # Normalize and simmetrize score
    df_hits= normalize_table(df_hits, mode)
    if not (gene_2_OG is None):
        df_hits= filter_hits_within_orthogroup(df_hits, gene_2_OG)
    # Select best hits
    return select_best_hits(df_hits, f_value)

def filter_hits_within_orthogroup(df_hits, gene_2_OG):
    F0= lambda row: (row.Query_accession in gene_2_OG) and (row.Target_accession in gene_2_OG)
    F1= lambda row: gene_2_OG[row.Query_accession]==gene_2_OG[row.Target_accession]
    F2= lambda row: F0(row) and F1(row)

    print('-----------------------------------')
    print(gene_2_OG)
    print(df_hits.apply(F0, axis= 1).sum())
    print(df_hits.apply(F1, axis= 1).sum())
    print(df_hits.apply(F2, axis= 1).sum())
    print('-----------------------------------')

    df_hits= df_hits[ df_hits.apply(F2, axis= 1) ].copy()
    columns= list(df_hits.columns)
    df_hits['OG']= df_hits.Query_accession.map(gene_2_OG)
    return df_hits[ ['OG']+columns ]

def pair_all_hits_raw(X, path):
    species_list,file_list= X
    # Load alignment hits
    F1= [path+x for x in file_list]
    return parse_prt_hits(F1, species_list)

def simple_species_parser(file):
    x0= file.find('.vs.')
    x1= x0+4
    x2= file.find(ext_alignment_hits)
    S0= file[:x0]
    S1= file[x1:x2]
    S1= '.'.join(S1.split('.')[:-1])
    return (S0,S1)

def select_best_hits(df_hits, f_value, a= 'Query_accession', b= 'Target_accession'):
    """
    Input:
    - path: A directory containing reciprocal blast-like analisis between pairs of species.
            The files contained there should be named as follows: <species_a>.fa.vs<species_b.fa>.blast
    - f_value: used for the dinamic threshold.
    """
    # Identify best hit scores w.r.t. query_gene and target_species
    w_x_B= max_hit_scores(df_hits)
    # Apply dynamic threshold for best hit selection
    #print('Selecting best hits by dynamic threshold...')
    is_best_hit= lambda row: row.Normalized_bit_score >= f_value * w_x_B[ row.Query_accession, row.Target_species ]
    mask= df_hits.apply(is_best_hit, axis= 1)

    #print('\n\n\n\n\n')
    #print(df_hits)
    #print(mask)
    #print('\n\n\n\n\n')

    df_best_hits= df_hits.loc[mask]
    return df_best_hits

def identify_orthogroups(df_best_hits,
                         Query_accession= 'Query_accession',
                         Target_accession= 'Target_accession',
                         Query_species= 'Query_species',
                         Target_species= 'Target_species',
                         N_max= 2000):
    # Create graph
    G= nx.DiGraph()
    G.add_edges_from( df_best_hits[[Query_accession,Target_accession]].values )
    # Identify connected components
    CCs= pd.Series(list(nx.weakly_connected_components(G)))
    # Reduce giant connected components
    mask_giant= CCs.apply(lambda X: len(X)>N_max)
    reduced_CCs= CCs[ mask_giant ].apply(lambda X: reduce_graph(nx.induced_subgraph(G, X), N_max= N_max))
    # Obtain singletons list
    singletons= reduced_CCs.str[1]
    singletons= set(chain.from_iterable(singletons.values))
    # Obtain graph partitions
    reduced_CCs= reduced_CCs.str[0]
    reduced_CCs= pd.Series(list(chain.from_iterable(reduced_CCs.values)), dtype='object')
    # Make final CCs list
    CCs= pd.concat((CCs[ ~mask_giant ], reduced_CCs), ignore_index= True)
    # Sort connected components by size and gene names
    CCs= pd.DataFrame(dict(CCs= CCs,
                           N= CCs.str.len(),
                           name= CCs.apply(lambda X: ''.join(sorted(''.join(X)))),
                          ))
    CCs= CCs.sort_values(by= ['N', 'name']).CCs
    # Asign a number for each connected component
    idx_2_og= dict(enumerate(CCs.values))
    D= {x: i for i,X in idx_2_og.items() for x in X}
    # Quit singletons w.r.t partition
    F= lambda x: x not in singletons
    df_best_hits= df_best_hits[ df_best_hits[Query_accession].apply(F) & df_best_hits[Target_accession].apply(F) ]
    # Keep only intra-hits w.r.t partition
    df_best_hits= df_best_hits[ df_best_hits.apply(lambda X: D[X[Query_accession]]==D[X[Target_accession]], axis= 1) ]

    # Asign a connected component to each row of dataframe
    cols= list(df_best_hits.columns)
    df_best_hits['OG']= df_best_hits[Query_accession].map(D)
    # Sort rows by OG
    df_best_hits= df_best_hits[['OG']+cols].sort_values('OG')

    # Create OGs table
    OGs_table= create_og_table(df_best_hits, idx_2_og,
                               Query_accession,
                               Target_accession,
                               Query_species,
                               Target_species)

    return df_best_hits, OGs_table

def identify_singletons(df, fastaspath):
    """
    df: dataframe for orthogroups
    fastaspath: list of fasta files
    """
    # Identify genes in orthogroups
    species_columns= list(df.columns)[3:]
    #F= lambda X: X.split(',')
    genes_og= df[species_columns].apply(lambda row: chain.from_iterable(row.dropna()) , axis= 1)
    #genes_og= df[species_columns].apply(lambda row: chain.from_iterable(F(X) for X in row.dropna()) , axis= 1)
    genes_og= set(chain.from_iterable(genes_og.values))

    # Identify genes in fasta files
    genes_fa= set()
    DD= dict()
    for file in fastaspath:
        with open(file) as F:
            fgenes= [title.split()[0] for title, _ in SimpleFastaParser(F)]
        genes_fa.update(fgenes)
        DD.update({x:file.split('/')[-1] for x in fgenes})
    singletons= genes_fa - genes_og
    singletons= [[x, DD[x]] for x in singletons]
    singletons= pd.DataFrame(singletons, columns= ['gene', 'file'])
    return singletons.sort_values('file')

def get_species_columns(genes, species, gene2species, idx):
    ret= {}
    for gene in genes:
        spec= gene2species[gene]
        ret[spec]= ret.get(spec, []) + [gene]

    nGenes= len(genes)
    nSpecies= len(ret)

    return [idx, nGenes, nSpecies] + [set(ret.get(spec, []))
                                      for spec in species]
    #return [idx, nGenes, nSpecies] + [','.join(ret.get(spec, []))
    #                                  for spec in species]

def create_og_table(df_BHs, idx_2_og,
                    Query_accession,
                    Target_accession,
                    Query_species,
                    Target_species):
    # Create genesdict
    gene2species= pd.DataFrame(
        list(df_BHs[[Query_accession,
                     Query_species]].values) + list(df_BHs[[Target_accession,
                                                            Target_species]].values
                                                   ),
        columns= ['gene', 'species']
    ).drop_duplicates().set_index('gene').species

    # Create species set
    species_list= sorted(set(gene2species))

    # Create map OG->species->gene
    OGs_table= pd.DataFrame([get_species_columns(genes, species_list, gene2species, idx)
                             for idx, genes in idx_2_og.items()],
                            columns= ['OG', 'n_genes', 'n_sepecies'] + species_list
                           )
    return OGs_table

def normalize_table(df_hits, normalization_method):
    # Normalize score
    normalized_score= normalize_scores(df_hits, normalization_method)
    F= lambda X: normalized_score[frozenset(X)]
    df_hits['Normalized_bit_score']= df_hits[['Query_accession', 'Target_accession']].apply(F, axis= 1)
    return df_hits

def normalize_scores(df, method):
    if method == 'query':
        normalized_score= df.Bit_score/df.Query_length
    elif method == 'target':
        normalized_score= df.Bit_score/df.Target_length
    elif method == 'alignment':
        normalized_score= df.Bit_score/df.Alignment_length
    elif method == 'smallest':
        normalized_score= df.Bit_score/df[['Query_length', 'Target_length']].min(axis=1)
    elif method == 'biggest':
        normalized_score= df.Bit_score/df[['Query_length', 'Target_length']].max(axis=1)
    elif method == 'raw':
        normalized_score= df.Bit_score
    elif method == 'Pdist':
        normalized_score= df.Sequence_identity.apply(lambda x:1-(x)/100)
    elif method == 'Pdist-JC':
        normalized_score= df.Sequence_identity.apply(calculate_jukes_cantor_correction)
    elif method == 'Pdist-LG':
        normalized_score= df.Sequence_identity.apply(lambda x:calculate_gamma_correction(x,1.0))
    elif method == 'Pdist-Dayoff':
        normalized_score= df.Sequence_identity.apply(lambda x:calculate_gamma_correction(x,2.25))
    else:
        raise ParameterError(f'Unrecognized method for normalization "{method}"')
    return symmetrize_scores(df, normalized_score, 'Query_accession', 'Target_accession')

def symmetrize_scores(df_hits, scores, a, b):
    # Map directed hit -> score
    DD= {(Q,T):S for Q,T,S in zip(df_hits[a], df_hits[b], scores)}
    # Set of undirected hits
    hits= {frozenset(X) for X in DD}
    # Map undirected hit -> score
    D1= {X : average_hit_scores(X,DD) for X in hits}
    return pd.Series(D1)

def average_hit_scores(X,DD):
    Q,T= X
    N=0
    S= 0
    edges= [(Q,T), (T,Q)]
    for edge in edges:
        if edge in DD:
            S+= DD[edge]
            N+= 1
    return S/N

def calculate_jukes_cantor_correction(seq_id):
    p_distance = 1-seq_id/100
    if p_distance == 0:
        return 0.0
    else:
        corrected_distance = - 19/20* log(1 - (20/19) * p_distance)
        return corrected_distance

def calculate_gamma_correction(seq_id, gamma_value):
    p_distance = 1-seq_id/100
    if p_distance == 0:
        return 0.0
    else:
        corrected_distance = gamma_value * ((1 - p_distance)**(-1/gamma_value)-1)
        return corrected_distance

pd_params= dict(names= _diamond_tl_headers, sep= '\t')
def parse_prt_hits(file_list, species_list, pd_params= pd_params):
    """
    Returns a DataFrame of hits obtained from proteinortho files
    """
    df_0= _read_bh_table(file_list[0], species_list, pd_params)
    df_1= _read_bh_table(file_list[1], species_list[::-1], pd_params)
    df= pd.concat((df_0, df_1))
    return df

def _read_bh_table(file, species, pd_params):
    df= pd.read_csv(file, **pd_params)
    columns= list(df.columns)
    df['Query_species']= species[0]
    df['Target_species']= species[1]
    return df[ ['Query_species', 'Target_species']+columns ]

def max_hit_scores(df_hits,
                   Query_accession= 'Query_accession',
                   Target_species= 'Target_species',
                   Score= 'Normalized_bit_score',
                  ):
    return df_hits.groupby([Query_accession, Target_species])[Score].max()



from rpy2 import robjects

def read_gene_trees(args, prefix, customEventAttr):
    df_rec= read_csv(f'{prefix}reconciliation.tsv', sep= '\t',  dtype= {'OG':str})
    # Mask gene trees
    if args.OGs_mask:
        df_rec= df_rec[ df_rec.OG.isin(args.OGs_mask) ]
    # Newick to nxTree
    df_rec.tree= df_rec.tree.apply(read_nhx)
    # Set mu map in proper format
    df_rec.reconciliation_map= df_rec.reconciliation_map.apply(txtMu_2_dict)
    # Set mu with nx node identifier
    df_rec.reconciliation_map= df_rec.apply(lambda row: update_mu(row.tree,
                                                    Ts, row.reconciliation_map,
                                                    'node_id', 'node_id'),
                                            axis=1)
    # Add custom event attr
    F= lambda T: add_eventAttr(T, 'label', customEventAttr)
    df_rec.tree.apply(F)
    return df_rec


desc_def= """
"""

class R_out:
    def __init__(self):
        self.capture_r_output()

    def capture_r_output(self):
        """
        Will cause all the output that normally goes to the R console,
        to end up instead in a python list.
        """
        # Import module #
        import rpy2.rinterface_lib.callbacks
        # Record output #
        self.stdout = []
        self.stderr = []
        # Dummy functions #
        def add_to_stdout(line): self.stdout.append(line)
        def add_to_stderr(line): self.stderr.append(line)
        # Keep the old functions #
        self.stdout_orig = rpy2.rinterface_lib.callbacks.consolewrite_print
        self.stderr_orig = rpy2.rinterface_lib.callbacks.consolewrite_warnerror
        # Set the call backs #
        rpy2.rinterface_lib.callbacks.consolewrite_print     = add_to_stdout
        rpy2.rinterface_lib.callbacks.consolewrite_warnerror = add_to_stderr

if __name__ == "__main__":
    from .plot_functions import get_tree_layoult, add_numbers, init_numbers, plot_dendogram, og_class_statistics, get_leaves_pos
    from .common_tools import norm_path, txtMu_2_dict, update_mu, add_eventAttr
    from .nhx_tools import read_nhx
    from .nxTree import is_leaf

    import argparse
    from importlib.metadata import version
    from pandas import read_csv
    import matplotlib.pyplot as plt
    from upsetplot import plot, UpSet

    from pandas import DataFrame, Series
    from collections import Counter
    from itertools import chain
    import seaborn as sns
    import matplotlib.patches as mpatches
    from math import ceil

    from tqdm import tqdm
    tqdm.pandas()

    V_tl= version('revolutionhtl')
    txt= f'REvolutionH-tl: Reconstruction of Evolutionaty Histories TooL (V{V_tl})'

    parser = argparse.ArgumentParser(prog= 'revolutionhtl',
                                     description=f'{txt}{desc_def}',
                                     usage='python -m revolutionhtl <arguments>',
                                     formatter_class=argparse.MetavarTypeHelpFormatter,
                                    )

    #############
    # Arguments #
    #############

    parser.add_argument('files_path',
                        help= '[str] Path of files outpued by REvolutionH-tl.',
                        type= norm_path,
                        nargs='?',
                        default= './',
                       )

    parser.add_argument('--files_prefix',
                        help= '[str] Prefix of files outpued by REvolutionH-tl.',
                        type= str,
                        nargs='?',
                        default= 'tl_project.',
                       )

    parser.add_argument('--species_order',
                        help= '[str] List of species in the desired order to plot.',
                        type= str,
                        nargs= '*',
                        default= None,
                       )

    parser.add_argument('--OGs_mask',
                        help= '[str] List of orthogroups IDs to mask gene trees at the file "gene_trees"',
                        type= str,
                        nargs= '*',
                        default= None,
                       )

    parser.add_argument('--output_path',
                        help= '[str | default: ./] Directory to save figures.',
                        type= norm_path,
                        default= './',
                       )

    parser.add_argument('--size',
                        help= '[float | default: 10] size of the side of the output figures',
                        type= float,
                        default= 10,
                       )

    parser.add_argument('--percentage_upsetplot',
                        help= '[int | default: 100] percentage of rows shown in upsetplot.',
                        type= int,
                        default= 100,
                       )

    args= parser.parse_args()

    from .hello import hello5
    print(f'{hello5}V{V_tl}\n')

    f_name= lambda name: f"{prefix}{name}.pdf"

    ####################
    # Define constants #
    ####################
    customEventAttr= 'customEventAttr'
    prefix= f'{args.files_path}{args.files_prefix}'
    s_tree_path_py= f'{prefix}labeled_species_tree.nhx'
    # Output parameters
    numbers_path= f'{prefix}gene_counts_reconciliation.tsv'

    #############
    # Load data #
    #############
    print("Loading data...")
    # Species tree
    #-------------
    with open(s_tree_path_py) as F:
        Ts= read_nhx(''.join( F.read().strip().split('\n') ))
    # Gene trees table
    #-----------------
    df_rec= read_gene_trees(args, prefix, customEventAttr)
    # Orthogroups
    #------------
    df_ogs= read_csv(f'{prefix}orthogroups.tsv', sep='\t').set_index('OG')
    # Species order and sorted leaves
    #--------------------------------
    if not args.species_order:
        _,_,_,aux= get_leaves_pos(Ts)
        args.species_order= [Ts.nodes[x]['label'] for x in aux]
    # Create a list of leaves sorted as specified by args.species_order
    label2node= {Ts.nodes[x]['label']:x for x in Ts if is_leaf(Ts,x)}
    sorted_leaves= [label2node[x] for x in args.species_order]
    # Sort columns of orthogroups dataframe
    df_ogs= df_ogs[ args.species_order ]
    # Count genes at each cell of orthogroups dataframe.
    from pandas import isna
    F= lambda X: 0 if isna(X) else len(X.split(','))
    df_ogs= df_ogs.map(F)

    # Singletones
    #------------
    df_sing= read_csv(f'{prefix}singletons.tsv', sep='\t')
    df_sing['species']= df_sing.file.str.split('.').str[0]
    singletones= df_sing.groupby('species').apply(lambda df: df.shape[0])
    # Orthologs
    #----------
    df_orth= read_csv(f'{prefix}orthologs.tsv', sep='\t')

    ################
    # Process data #
    ################
    print("Processing data...")

    # Species tree
    #-------------
    s_numbers= init_numbers(Ts)
    F= lambda row: add_numbers(row.tree, Ts, row.reconciliation_map, s_numbers)
    df_rec.progress_apply(F, axis= 1)

    df_numbers= [[Ts.nodes[vNode]['node_id'],
                  Dcounts['Dr'],
                  Dcounts['D'],
                  Dcounts['Sr'],
                  Dcounts['S'],
                  Dcounts['X'],
                 ]
                 for vNode,Dcounts in s_numbers.items() if vNode!=Ts.root]
    df_numbers= DataFrame(df_numbers, columns= ['node_id','duplication_roots','gene_gain_by_duplication','speciation_roots','genes_at_speciation','gene_loss'])
    df_numbers.to_csv(numbers_path,sep='\t',index=False)

    # UpSetPlot
    #----------
    class_statistics= og_class_statistics(df_ogs, args.species_order)

    nrows= class_statistics.shape[0]
    showRows= ceil( nrows*(args.percentage_upsetplot/100) )

    # Barplots
    #---------
    dropzero= lambda series: series[series!=0]

    # For each species classify orthogroups as single copy and multiple copy
    species_is_present= df_ogs>0
    OG_is_single_copy= df_ogs.apply(lambda row: (dropzero(row)==1).all(), axis=1)
    class_single_copy= DataFrame({'Single copy' : species_is_present.apply( lambda X: X & OG_is_single_copy ).sum(),
                                     'Multi copy' : species_is_present.apply( lambda X: X & ~OG_is_single_copy ).sum(),
                                    })
    class_single_copy.index.name= 'species'

    # For each species classify orthogroups by the number of species present
    count_species= lambda X: dropzero(X).shape[0]
    mask_sp= lambda species: df_ogs.loc[species_is_present[species]]
    class_n_species= DataFrame({species:Counter(mask_sp(species).apply(count_species, axis=1))
                                for species in df_ogs.columns})
    class_n_species.index.name= 'n_species'
    class_n_species.columns.name= None
    class_n_species= class_n_species.T
    class_n_species.index.name= 'species'
    class_n_species['Singletones']= singletones

    # For each species classify genes by the number of orthologs
    def get_n_orthologs_per_gene(df):
        df= DataFrame(chain(df[['species_a','a']].values,  df[['species_b','b']].values),
                      columns= ['species','gene'])
        df['one']= 1
        df= df.groupby('species')[['gene', 'one']].apply( lambda df: df.groupby('gene').one.sum().values ).to_dict()
        df= DataFrame([[species,x] for species,X in df.items() for x in X],
                      columns= ['species','n'])
        return df.set_index('species').n
    class_n_orthologs= get_n_orthologs_per_gene(df_orth)

    # For each species classify orthogroups as 1-1, 1-n, n-1, n-m
    species_is_single_copy= df_ogs==1
    species_is_1_n= lambda species: (species_is_single_copy[species] & ~(species_is_single_copy.drop(species, axis=1).all(axis=1)) ).sum()
    species_is_n_1= lambda species: (~species_is_single_copy[species] & species_is_single_copy.drop(species, axis=1).all(axis=1) ).sum()
    species_is_n_m= lambda species: (~species_is_single_copy[species] & ~(species_is_single_copy.drop(species, axis=1).all(axis=1)) ).sum()

    class_nm_orthogroups= DataFrame({'1-1' : species_is_present.apply( lambda X: X & OG_is_single_copy ).sum(),
                                     '1-n' : {species:species_is_1_n(species) for species in df_ogs.columns},
                                     'n-1' : {species:species_is_n_1(species) for species in df_ogs.columns},
                                     'n-m' : {species:species_is_n_m(species) for species in df_ogs.columns},
                                    })
    class_nm_orthogroups.index.name= 'species'



    ################
    # Plot figures #
    ################

    # Species tree
    #-------------
    fig, ax= plt.subplots(1,1, figsize= (3*args.size,args.size))
    plot_dendogram(Ts, s_numbers, ax, sorted_leaves= sorted_leaves)
    ax.axis('off')

    filename_tree = f_name("species_tree")
    fig.savefig(filename_tree, format="pdf", bbox_inches="tight")
    print(f'Written to {filename_tree}')

    width, height = fig.get_size_inches()
    uniform_figsize = (width, height)

    # UpSet plot
    #-----------
    upset= UpSet(class_statistics, sum_over='No. OGs',
                 sort_categories_by='input',
                 sort_by= '-cardinality',
                 intersection_plot_elements= 0,
                 orientation= 'vertical',
                 max_subset_rank= showRows,
                )

    upset.plot()

    fig= plt.gcf()
    fig.set_size_inches(3*args.size,args.size)
    for ax in fig.axes:
        ax.set_xticklabels([])
        ax.set_yticklabels([])

    filename_upset = f_name("upset_plot")
    fig.savefig(filename_upset, format="pdf", bbox_inches="tight")
    print(f'Written to {filename_upset}')

    # Barplots
    #---------
    fig, axs= plt.subplots(4,1,sharex=True, figsize= (uniform_figsize))
    class_single_copy.plot.bar(stacked=True, ax= axs[0], rot=0, legend='reverse')
    class_n_species.plot.bar(stacked=True, ax= axs[1], rot=0, legend='reverse')
    axs[2].violinplot([class_n_orthologs[species].values for species in df_ogs.columns], positions= list(range(len(df_ogs.columns))))
    class_nm_orthogroups.plot.bar(stacked=True, ax= axs[3], rot=0, legend='reverse')

    for ax in axs:
        ax.set_xticklabels([])
        ax.set_yticklabels([])

    axs[0].set_ylabel('Orthogroups')
    axs[1].set_ylabel('Genes')
    axs[2].set_ylabel('Genes')
    axs[3].set_ylabel('Orthogroups')
    axs[0].set_title('Single vs multy copy')
    axs[1].set_title('Species per orthogroup')
    axs[2].set_title('Orthologs per gene')
    axs[3].set_title('Orthologs and coorthologs')

    filename_barplots = f_name("barplots")
    fig.savefig(filename_barplots, format="pdf", bbox_inches="tight")
    print(f'Written to {filename_barplots}')

####

    from PyPDF2 import PdfReader, PdfWriter

    pdf_writer = PdfWriter()

    pdf_files = [filename_tree, filename_barplots, filename_upset]


    for pdf_file in pdf_files:
        reader = PdfReader(pdf_file)
        for page_num in range(len(reader.pages)):
            page = reader.pages[page_num]
            pdf_writer.add_page(page)

    outName= f_name('combined_plot')
    with open(outName, 'wb') as f:
        pdf_writer.write(f)
    print(f'Written to {outName}')

####

    ##################
    # Plot R summary #
    ##################
    from .r_plot_summary import modules

    # Run R modules
    rout= R_out()
    robjects.r( modules.Module_1 )
    robjects.r(f'species_tree_file <- "{s_tree_path_py}"')
    robjects.r(f'numbers_file <- "{numbers_path}"')
    robjects.r( modules.Module_2 )
    robjects.r( modules.Module_3 )
    robjects.r( modules.Module_4 )
    robjects.r( modules.Module_5 )
    robjects.r( f'o_format <- "pdf"' )
    robjects.r( f'prefix <- "{prefix}"' )
    robjects.r( modules.Module_6 )

    #######
    # End #
    #######



from .nxTree import is_leaf, get_children, get_dad, BFS_graph
from .error import notAllowedAttributeValue
from networkx import dfs_postorder_nodes, dfs_preorder_nodes
import matplotlib.pyplot as plt
from upsetplot import from_memberships
from pandas import DataFrame

##################
# Main functions #
##################
def plot_recon(Tg, Ts, mu,
               eventAttr= 'event',
               eventSymbols= {'S':'.r', 'D':'db'},
               sColor= '#97a4bd',
               gLineStyle= '-k',
               sRoot= 0,
               leafLabelAttr= 'label',
               leafLabelStyle= {'fontsize': 10, 'color': '#B37700'},
               pipe_r= 2,
               ax= None,
              ):
    # Compute graphics
    pos, branch_w, branch_d= get_tree_layoult(Ts, branch_r= pipe_r)
    Xpipe, Ypipe= get_pipe_layoult(Ts, pos, branch_w, sRoot)
    gPos= get_gTree_layoult(Tg, Ts, mu, pos, branch_w, eventAttr)

    # Plot
    if ax==None:
        fig, ax= plt.subplots(1,1,figsize=(6,6))

    Dsymbols= lambda x: eventSymbols.get(Tg.nodes[x][eventAttr], '.k')

    ax.fill(Xpipe, Ypipe, sColor)
    for vNode in filter(lambda x: (x!=Ts.root)and(not is_leaf(Ts,x)), Ts):
        x_mid, y= pos[vNode]
        x0,x1= x_mid-(branch_w/2),x_mid+(branch_w/2)
        ax.plot([x0,x1],[y,y],'gray')

    for xNode in dfs_postorder_nodes(Tg,Tg.u_lca):
        X= [x[0] for x in gPos[xNode]]
        Y= [x[1] for x in gPos[xNode]]

        if mu[xNode]==mu[get_dad(Tg,xNode)]:
            X= [X[0],X[-1]]
            Y= [Y[0],Y[-1]]

        ax.plot(X,Y,gLineStyle)

    for xNode in dfs_postorder_nodes(Tg,Tg.u_lca):
        x0= gPos[xNode][0][0]
        y0= gPos[xNode][0][1]
        ax.plot(x0,y0,Dsymbols(xNode))
        if is_leaf(Tg, xNode) and Tg.nodes[xNode][eventAttr]!='X':
            ax.text(x0,y0,f"{Tg.nodes[xNode].get(leafLabelAttr)}",
                    fontdict= leafLabelStyle,
                    rotation= -30,
                    rotation_mode= 'anchor',
                   )

    ax.axis('off')
    return ax


##############################################
# Back end functions for reconciliation plot #
##############################################

def get_leaves_pos(T, leaves= None, length= 1, r= 1, root= 0):
    """
    X-position of leaves in [0,length] with r = (branch width)/(branch separation)

    Arguments
    ---------
    T      : Tree
    leaves : Leaves sorted in the desired order
    length : Length of x-range to place leafs.
    r      : ratio (branch width)/(branch separation)

    Output
    ------
    lPos : Map node -> x-position
    w    : Branch width
    d    : Branch separation
    """
    if leaves==None:
        leaves= list(filter( lambda x: is_leaf(T,x), dfs_postorder_nodes(T, root) ))
    # Compute branch width and branch separation (assume length=1)
    n= len(leaves)
    d= 1/(n*r +n-1)
    w= r*d
    # Init output
    lPos= dict()
    i= w/2
    # Compute positions
    for x in leaves:
        lPos[x]= i * length
        i+= w+d
    return lPos, w, d, leaves

def get_tree_layoult(T, sorted_leaves= None, x_length= 1, y_length= 1, branch_r= 1, root= 0):
    """
    Return node position for a tree plot.

    The output layoult is the skeleton of a pipe-like tree
    where the branch-with and the branch-separation is defined by
    the argument branch_r = (branch width)/(branch separation).

    Arguments
    ---------
    T: Tree

    Keyword arguments
    -----------------
    sorted_leaves: list or tuple in the desired order of the leaves
    x_length     : x-length of the plot 
    y_length     : y-length of the plot 
    branch_r     : ratio (branch width)/(branch separation)
    root         : root of the tree for the plot

    Output
    ------
    pos  : Map node -> (x,y) position
    branch_w    : Branch width
    branch_d    : Branch separation
    """
    # Get leaf x-position
    leafs_pos, branch_w, branch_d,_= get_leaves_pos(T, leaves= sorted_leaves, length= x_length, r= branch_r)
    # Init output
    pos= dict()
    maxLevel= 0
    minX= dict()
    maxX= dict()
    # Compute output
    for uNode in dfs_postorder_nodes(T, root):
        if is_leaf(T, uNode):
            pos[uNode]= (leafs_pos[uNode], 0)
            minX[uNode]= maxX[uNode]= leafs_pos[uNode]
        else:
            # Compute position for node
            children= get_children(T, uNode)
            minX[uNode]= min((minX[ch] for ch in children))
            maxX[uNode]= max((maxX[ch] for ch in children))
            xu= (minX[uNode]+maxX[uNode])/2
            yu= max((pos[ch][1] for ch in children))+1
            pos[uNode]= (xu, yu)
            # update max level
            maxLevel= max((maxLevel, yu))
    # Normalize height
    norm= maxLevel/y_length
    pos= {x:(y[0], y[1]/norm) for x,y in pos.items()}
    return pos, branch_w, branch_d

def get_pipe_layoult(T, pos, branch_w, root= 0):
    """
    Return the perimeter of the pipe of a tree.

    Arguments
    ---------
    T        : Tree
    pos      : Map node -> (x,y) position. This is the skeleton of the pipe
    branch_w : Branch width

    Keyword arguments
    -----------------
    root : root of the tree for the plot

    Output
    ------
    Xpipe : List of x-positions of the pipe
    Ypipe : List of y-positions of the pipe
    """
    # Constant to compute spesiation interval
    w2= branch_w/2
    # Init output
    Xpipe= dict()
    Ypipe= dict()
    # compute output
    for uNode in dfs_postorder_nodes(T, root):
        # compute speciation interval of uNode
        x_u= pos[uNode][0]
        x0_u, x1_u= x_u-w2, x_u+w2
        y_u= pos[uNode][1]
        # compute pipe of uNode
        if is_leaf(T, uNode):
            Xpipe[uNode]= [x0_u, x1_u]
            Ypipe[uNode]= [y_u, y_u]
        else:
            uChildren= list(get_children(T, uNode))
            # Init pipe for uNode
            Xpipe[uNode]= [x0_u]
            Ypipe[uNode]= [y_u]
            # Add pipe of children at the left side of a bifurcation
            for v0Node,v1Node in consecutive_children_pairs(uChildren):
                # Compute point of bifurcation
                x1_v0, y_v0= Xpipe[v0Node][-1], Ypipe[v0Node][0]
                x0_v1, y_v1= Xpipe[v1Node][0], Ypipe[v1Node][0]
                xMid, yMid= midpoint(x0_u, x1_u, y_u, x1_v0, y_v0, x0_v1, y_v1)
                # Update pipe
                Xpipe[uNode]+= Xpipe[v0Node] + [xMid]
                Ypipe[uNode]+= Ypipe[v0Node] + [yMid]
            # Add pipe of the last child
            v1Node= uChildren[-1]
            Xpipe[uNode]+= Xpipe[v1Node] + [x1_u]
            Ypipe[uNode]+= Ypipe[v1Node] + [y_u]
            # Delete pipe of children
            for vNode in uChildren:
                del Xpipe[vNode]
                del Ypipe[vNode]

    return Xpipe[root], Ypipe[root]

def get_gTree_layoult(Tg, Ts, mu, sPos, branch_w, eventAttr):
    """
    Return the layoult of a reconciliated gene tree inside a species tree layoult

    Arguments
    ---------
    Tg        : Gene tree
    Ts        : Species tree
    mu        : Map gNode -> sNode. Reconciliation map
    sPos       : Map sNode -> (x,y) position. Species tree layoult
    branch_w  : Branch width of species tree layoult
    eventAttr : Attribute in gene tree indicating evolutionary event

    Output
    ------
    gPos
    """

    node_level, D_branch_roots= compute_levels_in_branch(Tg, Ts, mu, Tg.u_lca, eventAttr)
    y_gPos= get_y_pos(Tg, Ts, mu, sPos, eventAttr, node_level)
    y_cumulative, eMap, y_spots= count_genes(Tg, Ts, mu, y_gPos, D_branch_roots, node_level, eventAttr)
    gPos= get_xy_pos(Tg, y_gPos, Ts, sPos, y_spots, y_cumulative, mu, D_branch_roots, eMap, branch_w)
    return gPos

def get_xy_pos(Tg, y_gPos, Ts, sPos, y_spots, y_cumulative, mu, D_branch_roots, eMap, branch_w):
    idxNext= {X:0 for X,count in y_cumulative.items()}
    gPos= {Tg.root:[sPos[mu[Tg.root]]]}
    for uNode in dfs_preorder_nodes(Tg, Tg.u_lca):
        vNode= mu[uNode]
        spot= y_gPos[uNode]
        gPos[uNode]= get_branch_root_edge(uNode, vNode, spot, eMap, Ts, sPos, idxNext, y_spots, branch_w, y_cumulative, gPos, Tg, D_branch_roots)
    return gPos

def get_branch_root_edge(uNode, vNode, spot, eMap, Ts, sPos, idxNext, y_spots, branch_w, y_cumulative, gPos, Tg, D_branch_roots):
    getX= lambda vNode,spot: set_x_pos(Ts, vNode, spot, sPos, idxNext, branch_w, y_cumulative)
    ret= []

    top= gPos[get_dad(Tg, uNode)][0]
    top_x,top_y= top
    F= lambda x: top_y>x>=spot

    for vUpper in eMap[uNode]:
        ret+= [(getX(vUpper,Spot1), Spot1) for Spot1 in filter(F, y_spots[vUpper])]
    if uNode!=Tg.u_lca:
        ret+= [top]
    return ret

def set_x_pos(Ts, vNode, spot, sPos, idxNext, branch_w, y_cumulative):
    X= (vNode,spot)
    x0,y0= sPos[vNode]
    x1,y1= sPos[get_dad(Ts, vNode)]
    if x0==x1:
        x_s= x0
    else:
        m= (y1-y0)/(x1-x0)
        b= y0 - m*x0
        x_s= (spot-b)/m
    idxNext[X]+= 1
    return x_s - (branch_w/2) + ( idxNext[X]*branch_w / (y_cumulative[X]+1) )

def count_genes(Tg, Ts, mu, gPos, D_branch_roots, node_level, eventAttr):
    """
    For each branch of the species tree return:
    - y-axis spots over the brach where nodes of gene tree has been mapped, and
    - the cumulative number of genes in the species as the y-axis decreases.
    Additionally, for each edge uv of the gene tree return:
    - The sequense of node S of the species tree such that mu[v] < S <= mu[u]
    """
    y_cumulative= dict()
    y_spots= dict()

    eMap= {uNode : edge_map(uNode, Tg, Ts, mu) for uNode in Tg}

    for vNode in BFS_graph(Ts, Ts.u_lca):
        # Track duplication of genes w.r.t. a spot in parental branch of vNode
        branch_roots= D_branch_roots[vNode]
        spot_2_increase= dict()
        for uNode in branch_roots:
            for uPrime in branch_bfs(Tg, uNode, mu):
                spot= gPos[uPrime]
                if node_level[uPrime] > 1:
                    spot_2_increase[spot]= spot_2_increase.get(spot, 0) + len(Tg[uPrime]) - 1
                else:
                    spot_2_increase[spot]= spot_2_increase.get(spot, 0)

        # Count total genes w.r.t y spot
        y_cumulative[vNode]= dict() # Map : spot -> number of genes just befor a time, i.e. at the minimum time > spot
        y_spots[vNode]= sorted(spot_2_increase)
        #cumulative_count= sum(map(lambda uNode: node_level[uNode]>1, branch_roots))
        cumulative_count= len(branch_roots)
        for spot in y_spots[vNode][::-1]:
            y_cumulative[vNode][spot]= cumulative_count
            cumulative_count+= spot_2_increase[spot]
        # Track the y positions of parental edges
        for uNode in branch_roots:
            bottom= gPos[uNode]
            #top= gPos[get_dad(Tg, uNode)]
            top= y_spots[vNode][-1]
            F= lambda x: top>x>bottom
            for vUpper in eMap[uNode]:
                for spot in filter(F, y_spots[vUpper]):
                    y_cumulative[vUpper][spot]+= 1

    # Flatten output
    y_cumulative= {(vNode,spot) : count
                   for vNode,spot2count in y_cumulative.items()
                   for spot,count in spot2count.items()}

    return y_cumulative, eMap, y_spots

def edge_map(uNode, Tg, Ts, mu):
    """
    Tracks the position of the parental edge of uNode in parental branches of the species tree
    """
    if uNode==Tg.root:
        return []
    uUpper= get_dad(Tg, uNode)
    vNode= mu[uNode]
    vUpper= mu[uUpper]
    if vNode==vUpper:
        return [vNode]
    ret= []
    while vNode!=vUpper:
        ret+= [vNode]
        vNode= get_dad(Ts, vNode)
    return ret

def branch_bfs(Tg, xNode, mu):
    queue= [xNode]
    idx= 0
    n= 1
    while idx<n:
        xNode= queue[idx]
        ch= list(branch_descendants(Tg, xNode, mu))
        queue+= ch
        n+= len(ch)
        idx+= 1
    return queue


def branch_descendants(Tg, xNode, mu):
    yNode= mu[xNode]
    return filter(lambda xPrime: mu[xPrime]==yNode, Tg[xNode])

def get_y_pos(Tg, Ts, mu, sPos, eventAttr, node_level):
    """
    y-axis position of all the nodes of the gene tree
    """
    gPos= {Tg.root:sPos[Ts.root][1]}
    for uNode in BFS_graph(Tg, Tg.u_lca):
        # See current node
        event= Tg.nodes[uNode][eventAttr]
        dad_u= get_dad(Tg,uNode)
        # map of current node
        vNode= mu[uNode]
        dad_v= get_dad(Ts, vNode)
        y_pos_v=  sPos[vNode][1]
        # Compute spot in branch
        if is_leaf(Tg, uNode) and event!='X':
            spot= 0
        elif event=='S':
            spot= y_pos_v
        elif event in ['D', 'X']:
            if vNode==mu[dad_u]:
                up= gPos[dad_u]
            else:
                up= sPos[dad_v][1]
            spot= y_pos_v + (up-y_pos_v)*(1-(1/node_level[uNode]))
        else:
            raise notAllowedAttributeValue(f'Unknown event "{event}"')
        gPos[uNode]= spot
    return gPos

def compute_levels_in_branch(Tg, Ts, mu, gRoot, eventAttr):
    """
    For each branch of the species tree return:
    - level of nodes in the sub-gene trees restructed to the branch of the species tree
    - the roots of the sub-gene trees restructed to the branch of the species tree

    Note:
    Each edge uv is represented by the child node v.
    The level 0 corresponds to genes mapping to the speciation event v.
    """
    node_level= {}
    D_branch_roots= {y:list() for y in Ts}
    for xNode in dfs_postorder_nodes(Tg, gRoot):
        event= Tg.nodes[xNode][eventAttr]
        yNode= mu[xNode]
        if yNode!=mu[get_dad(Tg, xNode)]:
            D_branch_roots[yNode].append(xNode)
        if is_leaf(Tg, xNode):
            if event=='X': node_level[xNode]= 2
            else: node_level[xNode]= 1
        else:
            if event=='S': node_level[xNode]= 1
            elif event=='D': node_level[xNode]= compute_d_level(Tg, xNode, mu, node_level)
            else: raise notAllowedAttributeValue(f'Unknown event "{event}"')
    return node_level, D_branch_roots

def compute_d_level(Tg, xNode, mu, node_level):
    maxLevel= max([node_level[xPrime] for xPrime in branch_descendants(Tg, xNode, mu)]+[0])
    return maxLevel+1

###################################
# Functions for summary dendogram #
###################################

# Graphics
#---------
def plot_dendogram(T, numbers, ax, delta_txt=0.04, sorted_leaves=None):
    pos,_,_= get_tree_layoult(T, sorted_leaves= sorted_leaves)
    for u_node in dfs_postorder_nodes(T, 0):
        if is_leaf(T, u_node):
            plot_leaf(T, pos, u_node, numbers, ax, delta_txt)
        else:
            plot_inner(T, pos, u_node, numbers, ax, delta_txt)
    return None

def plot_leaf(T, pos, u_node, numbers, ax, delta_txt):
    x_u,y_u= pos[u_node]
    txt= f'{get_txt(u_node, numbers, "S")}'
    #txt= f"{numbers[u_node]['S']} ({numbers[u_node]['Sr']})"
    species= T.nodes[u_node]['label']
    ax.text(x_u, y_u, txt, c='red')
    ax.text(x_u, y_u-delta_txt, species, c='k', rotation=-30, rotation_mode= 'anchor',)
    return None

def get_txt(node, numbers, symbol, prefix=''):
    txt= ''
    if numbers[node][f'{symbol}']:
        txt+= f"{numbers[node][f'{symbol}']} "
    if numbers[node][f'{symbol}r']:
        txt+= f"({numbers[node][f'{symbol}r']})"
    if len(txt)>0:
        txt= prefix+txt
    return txt

def plot_inner(T, pos, u_node, numbers, ax, delta_txt):
    # Positions
    x_u,y_u= pos[u_node]
    ch= list(T[u_node])
    x0,x1= get_range(ch, pos)
    # Plot speciation line
    ax.plot([x0,x1],[y_u,y_u],'k')
    txt= get_txt(u_node, numbers, 'S')
    ax.text(x_u, y_u, txt, c='red')
    # Plot children
    for v_node in ch:
        x_v,y_v= pos[v_node]
        y_mid= (y_u+y_v)/2
        # Branch
        ax.plot([x_v,x_v], [y_u, y_v],'k')
        # Numbers
        txt= f"{get_txt(v_node, numbers, 'D', '+ ')}"
        ax.text(x_v, y_mid, txt, c='blue')
        txt= f"{get_txt(v_node, numbers, 'X', '- ')}"
        ax.text(x_v, y_mid-delta_txt, txt, c='green')
    return None

def get_range(ch, pos):
    ch_pos= set((pos[x][0] for x in ch))
    x0,x1= min(ch_pos), max(ch_pos)
    return x0,x1

init_numbers= lambda T: {x:{s:0 for s in ['S','D','X','Sr','Dr','Xr']} for x in T}

def get_symbol(T, x, gRoot):
    label= T.nodes[x]['label']
    if x==gRoot:
        label+= 'r'
    if is_leaf(T,x):
        if label=='X':
            return label
        return 'S'
    return label

def add_numbers(Tg, Ts, mu, numbers=None, gRoot= 1):
    if numbers==None:
        numbers= init_numbers(Ts)
    for x_node in dfs_postorder_nodes(Tg, 1):
        x_symbol= get_symbol(Tg, x_node, gRoot)
        y_node= mu[x_node]
        if x_symbol=='Dr':
            numbers[y_node][x_symbol]+= 1
            numbers[y_node]['D']+= len(Tg[x_node])-1
        elif x_symbol=='D':
            numbers[y_node][x_symbol]+= len(Tg[x_node])-1
        else:
            numbers[y_node][x_symbol]+= 1
    return numbers

###########################
# Functions for UpSetPlot #
###########################

# Classify OGs by precense of species and count
#----------------------------------------------
def og_class_statistics(df, species_list): # previous name of this function: def clade_info()
    """
    Classify orthogroups using the set of species represented by genes members of the orthogroup.
    Then returns numerical descriptions of the classes.

    Arguments
    ---------
    df           : Dataframe describing orthogroups.
                   column names are species, rows names are orthogroups, cells are set of genes.
    species_list : List of the species in df, sorted as desired in the visualization.

    Output
    ------
    class_statistics

    """
    # Classify orthogroups
    F= lambda row: list( row[[x>0 for x in row]].index )
    og_classification= from_memberships(df.progress_apply(F, axis= 1))
    og_classification= og_classification.reset_index().set_index(species_list).ones
    # Count genes per orthogroup
    n_genes_per_orthogroup= df.sum(axis=1)
    n_genes_per_orthogroup.index= og_classification.index
    # Count genes per species per orthogroup
    av_genes_per_speciesper_orthogroup= df.apply(lambda row: row[row>0].mean() , axis=1)
    av_genes_per_speciesper_orthogroup.index= og_classification.index
    # Group by class
    gb_n_genes_per_orthogroup= n_genes_per_orthogroup.groupby(species_list)
    # Compute class statistics
    class_statistics= DataFrame({
        'No. OGs' : og_classification.groupby(species_list).sum(),
        'No. genes' : gb_n_genes_per_orthogroup.sum(),
        'Genes per OG' : gb_n_genes_per_orthogroup.mean(),
        'Genes per species' : av_genes_per_speciesper_orthogroup.groupby(species_list).mean(),
    })
    return class_statistics

##########################
# Functions for barplots #
##########################


# Clades of the tree in binary code for upSetplot
#------------------------------------------------
def get_bool_clades(T):
    Il= {} # Induced leafs
    X= []
    for u_node in dfs_postorder_nodes(T, 1):
        if is_leaf(T, u_node):
            # Induced species
            leaf_s= T.nodes[u_node]['species']
            Il[u_node]= {leaf_s}
        else:
            ch= list(T[u_node])
            # Compute species in leaves
            Il[u_node]= set(chain.from_iterable((Il[v_node] for v_node in ch)))
            # Determine bool clade
            X+= [tuple(x in Il[u_node] for x in species_list)]
    return X

######################
# Auxiliar functions #
######################

def where_is_inherited(gNode, Tg, Ts, mu, eventAttr, gRoot):
    """
    Returns a list of tuples (sNode,event),
    the first element is the map of gNode,
    and the ancestors of (sNode,event)
    just before the map of gDad.
    """
    # Map of gNode
    X0= (mu[gNode], Tg.nodes[gNode][eventAttr])
    # If gNode is root, then it is mapped to only one element in S
    if gNode==gRoot:
        return [X0]
    # Otherwise, compute the upper limit
    gDad= get_dad(Tg, gNode)
    X1= (mu[gDad], Tg.nodes[gDad][eventAttr])
    # Return the ancestors in S before the upper limit
    if X0==X1:
        return [X0]
    ext_path= []
    while X0!=X1:
        ext_path+= [X0]
        X0= get_map_dad(X0, Ts)
    return ext_path

def get_map_dad(X0, Ts):
    sNode,event= X0
    if event=='S':
        return (sNode,'D')
    return (get_dad(Ts, sNode),'S')

def consecutive_children_pairs(ch):
    return zip(ch[0:], ch[1:])

def midpoint(x0_u, x1_u, y_u, x1_v0, y_v0, x0_v1, y_v1):
    m0= (y_u-y_v0)/(x1_u-x1_v0)
    m1= (y_u-y_v1)/(x0_u-x0_v1)

    xMid=(y_v1-y_v0 +m0*x1_v0-m1*x0_v1)/(m0-m1)
    yMid= m0*(xMid-x1_v0)+y_v0

    return xMid, yMid




from .nxTree import copy_tree, is_leaf, induced_colors
from .nhx_tools import read_nhx, tralda_to_nxTree, get_nhx
from .in_out import get_digraph_cols, write_digraphs_list

from numpy import argsort
from itertools import product, combinations
import networkx as nx
from bmgedit import BMGEditing
import pandas as pd
from math import ceil
from tqdm import tqdm
tqdm.pandas()


def build_graph(G,
                method='louvain',
                objective='cost',
                color_attr= None,
                use_binary_triples=True,
                force_binary= False,
                double_build= True,
               ):
    """
    ARG1: method - this param will be use to select the method
    """
    if color_attr != None:
        G1= G.copy()
        for x in G:
            G1.nodes[x]['color']= G.nodes[x][color_attr]
    else:
        G1= G

    solver = BMGEditing.BMGEditor(G1, binary= force_binary, use_binary_triples= use_binary_triples)
    solver.build(method, objective)

    hug_free= solver.get_bmg(extract_triples_first=False, update_tree= double_build)

    accession_2_species= {}
    nId_2_accession= {}
    for x in G:
        XX= G.nodes[x]['accession']
        YY= G.nodes[x]['species']
        hug_free.nodes[x]['accession']= XX
        hug_free.nodes[x]['species']= YY
        accession_2_species[XX]= YY
        nId_2_accession[x]= XX

    nxT= tralda_to_nxTree(solver._tree, name_attr= 'accession')
    # color_attr
    for x in nxT:
        if len(nxT[x])==0:
            nId= int(nxT.nodes[x].get('accession','NO_LABEL'))
            accession= nId_2_accession[nId]
            color= accession_2_species.get(accession)
            nxT.nodes[x]['species']= color
            nxT.nodes[x]['accession']= accession
        else:
            nxT.nodes[x]['species']= None

    return hug_free, nxT

def build_graph_series(G, args):
    lens= G.apply(len)
    is_sorted= all([lens.iloc[i] <= lens.iloc[i+1]
                    for i in range(lens.shape[0]-1)])
    if not is_sorted:
        order= argsort(lens.values)
        G= G.iloc[order]
        lens= lens.iloc[order]

    # Compute partition limits
    M= max(lens)
    f_low= lambda i: i*args.k_size_partition + 1
    f_up= lambda i: (i+1)*args.k_size_partition
    limits= ((f_low(i), f_up(i)) for i in range(0, ceil(M/args.k_size_partition)))
    total= len(G)
    Stotal= 0

    # File names
    bmg_opath= f'{args.output_prefix}.best_matches.tsv'
    with open(bmg_opath, 'w') as F:
        F.write('\t'.join(get_digraph_cols())+'\n')
    ors_opath= f'{args.output_prefix}.orthologs.tsv'
    with open(ors_opath, 'w') as F:
        F.write('a\tb\tOG\n')
    gtr_opath= f'{args.output_prefix}.BMG_gene_trees.tsv'
    with open(gtr_opath, 'w') as F:
        F.write('OG\ttree\n')

    # Start process
    Tg_list= []
    for down,up in limits:
        mask= lens.apply(lambda x: down <= x <= up)
        lMask= sum(mask)

        if lMask>0:

            Stotal+= lMask
            print('\n')
            print(f'| Reconstructing evolution for families with {down} <= number of genes <= {up}')
            print(f'| This represents the {100*lMask/total:.2f} % of the data')
            print(f'| At the end of this bunch, {100*Stotal/total:.2f} % of the data will be analyzed.\n')
            print('Reconstructing gene trees...')

            G_T= G[mask].progress_apply(lambda X: build_graph(X,
                                                              color_attr= 'species',
                                                              method= args.bmg_heuristic,
                                                              use_binary_triples= args.no_binary_triples,
                                                              force_binary= args.force_binary_gene_tree,
                                                              double_build= args.gene_tree_no_double_build,
                                                             ))

            G[mask]= G_T.apply(lambda X: X[0])
            Tg= G_T.apply(lambda X: X[1])
            Tg.name='tree'

            print('Labeling gene tree with evolutionary events...')
            Tg= Tg.progress_apply( lambda x: get_augmented_tree(x, 'accession', 'species') )
            Tg_list+= [Tg]
            Tgtxt= Tg.apply(lambda x: get_nhx(x, root= 1, name_attr= 'accession')).reset_index()

            write_digraphs_list(G[mask], bmg_opath, mode= 'a', header= False)
            print(f'Best match graphs successfully written to {bmg_opath}')
            Tgtxt.to_csv(gtr_opath, sep='\t', index= False, mode= 'a', header= False)
            print(f'Gene trees successfully written to {gtr_opath}')

    Tg= pd.concat(Tg_list)
    return Tg

##################################################################
##################################################################
##################################################################

def _addN(T,idx, tree, n):
    T.add_node(idx, **(tree.nodes[n].copy()))

def get_augmented_tree(tree, geneAttr, speciesAttr):
    """
    Constructs the color intersection graph
    Schaller et. al. 2020:
    Complete characterization of incorrect orthology assiments in BMGs
    """

    T= nx.DiGraph()
    T.root= 0
    _addN(T, 0, tree, 0)
    w_idx= 1
    mu= {1 : 0} # node x in tree --> corresponding dad in T

    for node in list(nx.dfs_preorder_nodes(tree, source= 1)):

        _addN(T, w_idx, tree, node)
        T.add_edge(mu[node], w_idx)
        for child in tree[node]:
            mu[child]= w_idx
        current= w_idx
        w_idx+= 1

        if (not is_leaf(tree, node)) and (node != tree.root):
            CIG= get_CIG(tree, node, speciesAttr)
            CC= list(nx.connected_components(CIG))
            if len(CC)>1:
                T.nodes[current][geneAttr]= 'S'
                for cc in (X for X in CC if len(X)>1):
                    _create_redundant_node(T, w_idx, current, tree, node, cc, mu, geneAttr, speciesAttr)
                    w_idx+= 1
            else:
                T.nodes[current][geneAttr]= 'D'
    T.nodes[T.root][geneAttr]= None
    return T

def get_CIG(tree, node, speciesAttr):
    """
    Color Intersection Graph
    """
    children= set(tree[node])
    CIG= nx.Graph()
    CIG.add_nodes_from(children)
    for v0,v1 in combinations(children, 2):
        colors0= induced_colors(tree, v0, speciesAttr)
        colors1= induced_colors(tree, v1, speciesAttr)
        if not colors0.isdisjoint(colors1):
            CIG.add_edge(v0, v1)
    return CIG

def _create_redundant_node(T, w_idx, current, tree, node, cc, mu, geneAttr, speciesAttr):
    T.add_node(w_idx, **{geneAttr: 'D', speciesAttr: None})
    T.add_edge(current, w_idx)
    for child in cc:
        mu[child]= w_idx
